import logging

from ...utils.service_discovery import find_service_url


class PrometheusDiscovery:
    prometheus_url = None
    @classmethod
    def find_prometheus_url(cls):
        """
        Try to autodiscover the url of an in-cluster grafana service
        """
        if cls.prometheus_url:
            return cls.prometheus_url
        prometheus_selectors = ["app=kube-prometheus-stack-prometheus", "app.kubernetes.io/name=prometheus"]
        for label_selector in prometheus_selectors:
            service_url = find_service_url(label_selector)
            if service_url:
                cls.prometheus_url = service_url
                return service_url
        logging.error("Prometheus url could not be found. Add 'prometheus_url' under global_config")
        return None



import logging
from functools import wraps

import hikaru
from hikaru.model import *

from ...core.model.k8s_operation_type import K8sOperationType
from ..base_handler import handle_event
from ...core.model.cloud_event import *
from ...core.model.playbook_hash import playbook_hash
from ...core.model.runner_config import *
from ...core.active_playbooks import register_playbook, activate_playbook
from pydantic import BaseModel

from ...core.model.events import EventType
from ...core.triggers import TriggerParams
from .autogenerated.events import KIND_TO_EVENT_CLASS, KubernetesAnyEvent
from .autogenerated.models import get_api_version
from .base_event import K8sBaseEvent
from .model_not_found_exception import ModelNotFoundException


class IncomingK8sEventPayload(BaseModel):
    """
    The format of incoming CloudEvent payloads containing kubernetes events. This is mostly used for deserialization.
    """

    operation: str
    kind: str
    apiVersion: str = ""
    clusterUid: str
    description: str
    obj: Dict[Any, Any]
    oldObj: Optional[Dict[Any, Any]]


def parse_kubernetes_objs(k8s_payload: IncomingK8sEventPayload):
    model_class = get_api_version(k8s_payload.apiVersion).get(k8s_payload.kind)
    if model_class is None:
        msg = f"classes for kind {k8s_payload.kind} cannot be found. skipping. description {k8s_payload.description}"
        logging.error(msg)
        raise ModelNotFoundException(msg)

    obj = hikaru.from_dict(k8s_payload.obj, cls=model_class)
    old_obj = None
    if k8s_payload.oldObj is not None:
        old_obj = hikaru.from_dict(k8s_payload.oldObj, cls=model_class)
    return (obj, old_obj)


def parse_incoming_kubernetes_event(
    k8s_payload: IncomingK8sEventPayload,
) -> Optional[K8sBaseEvent]:
    (obj, old_obj) = parse_kubernetes_objs(k8s_payload)
    event_class = KIND_TO_EVENT_CLASS.get(k8s_payload.kind)
    if event_class is None:
        logging.error(
            f"classes for kind {k8s_payload.kind} cannot be found. skipping. description {k8s_payload.description}"
        )
        return None
    operation_type = K8sOperationType(k8s_payload.operation)
    return event_class(
        operation=operation_type,
        description=k8s_payload.description.replace("\n", ""),
        obj=obj,
        old_obj=old_obj,
    )


def parse_incoming_kubernetes_any_event(
    k8s_payload: IncomingK8sEventPayload,
) -> Optional[K8sBaseEvent]:
    (obj, old_obj) = parse_kubernetes_objs(k8s_payload)
    operation_type = K8sOperationType(k8s_payload.operation)
    return KubernetesAnyEvent(
        operation=operation_type,
        description=k8s_payload.description.replace("\n", ""),
        obj=obj,
        old_obj=old_obj,
    )


def register_k8s_playbook(
    func,
    kind,
    operation: Optional[K8sOperationType],
    name_prefix="",
    namespace_prefix="",
):
    register_playbook(
        func,
        deploy_on_k8s_topology,
        TriggerParams(
            name_prefix=name_prefix,
            namespace_prefix=namespace_prefix,
            kind=kind,
            operation=operation,
        ),
    )
    return func


def register_k8s_any_playbook(
    func, operation: Optional[K8sOperationType], name_prefix="", namespace_prefix=""
):
    register_playbook(
        func,
        deploy_on_any_k8s_topology,
        TriggerParams(
            name_prefix=name_prefix,
            namespace_prefix=namespace_prefix,
            operation=operation,
        ),
    )
    return func


def prefix_match(prefix, field_value) -> bool:
    if prefix == "":
        return True
    if (
        field_value is None
    ):  # we have a prefix requirement, but field doesn't exist. no match
        return False
    return field_value.startswith(prefix)


def k8s_topology_handler(
    cloud_event: CloudEvent,
    event_parser,
    func,
    trigger_params: TriggerParams,
    named_sinks: List[str],
    action_params=None,
):
    logging.debug(f"checking if we should run {func} on k8s event")
    k8s_payload = IncomingK8sEventPayload(**cloud_event.data)

    if not k8s_topology_event_matched(k8s_payload, trigger_params):
        return "SKIP"

    try:
        concrete_event = event_parser(k8s_payload)
    except ModelNotFoundException:
        return "SKIP"

    return handle_event(func, concrete_event, action_params, "kubernetes", named_sinks)


def deploy_on_k8s_topology(
    func, trigger_params: TriggerParams, named_sinks: List[str], action_params=None
):
    @wraps(func)
    def wrapper(cloud_event: CloudEvent):
        return k8s_topology_handler(
            cloud_event,
            parse_incoming_kubernetes_event,
            func,
            trigger_params,
            named_sinks,
            action_params,
        )

    playbook_id = playbook_hash(func, trigger_params, action_params)
    activate_playbook(EventType.KUBERNETES_TOPOLOGY_CHANGE, wrapper, func, playbook_id)
    return wrapper


def deploy_on_any_k8s_topology(
    func, trigger_params: TriggerParams, named_sinks: List[str], action_params=None
):
    @wraps(func)
    def wrapper(cloud_event: CloudEvent):
        return k8s_topology_handler(
            cloud_event,
            parse_incoming_kubernetes_any_event,
            func,
            trigger_params,
            named_sinks,
            action_params,
        )

    playbook_id = playbook_hash(func, trigger_params, action_params)
    activate_playbook(EventType.KUBERNETES_TOPOLOGY_CHANGE, wrapper, func, playbook_id)
    return wrapper


def k8s_topology_event_matched(
    event: IncomingK8sEventPayload, trigger_params: TriggerParams
):
    if (
        trigger_params.operation is not None
        and trigger_params.operation.value != event.operation
    ):
        logging.debug(f"operation {event.operation} != {trigger_params.operation}")
        return False

    if trigger_params.kind is not None and trigger_params.kind != event.kind:
        logging.debug(f"kind {event.kind} != {trigger_params.kind}")
        return False

    metadata = event.obj.get("metadata", {})
    if trigger_params.name_prefix != "":
        obj_name = metadata.get("name", "")
        if not obj_name.startswith(trigger_params.name_prefix):
            logging.debug("name doesn't match")
            return False

    if trigger_params.namespace_prefix != "":
        obj_namespace = metadata.get("namespace", "")
        if not obj_namespace.startswith(trigger_params.namespace_prefix):
            logging.debug("namespace doesn't match")
            return False

    return True



import inspect
import logging
from typing import Callable, Dict


# TODO: could be extracted to a small library - see e.g. https://stackoverflow.com/questions/16576553/python-only-pass-arguments-if-the-variable-exists
def call_with_optional_params(func: Callable, available_args: Dict[str, object]):
    args_to_use = []  # must be in order
    expected_args = inspect.getfullargspec(func)
    for arg_name in expected_args.args + expected_args.kwonlyargs:
        if arg_name in available_args:
            args_to_use.append(available_args[arg_name])
        else:
            raise Exception(
                f"function requires argument that we don't recognize by name {arg_name}"
            )
    logging.info(
        f"available_args={available_args} expected_args={expected_args} args_to_use=f{args_to_use}"
    )
    return func(*args_to_use)



from enum import Enum
from typing import List, Any, Optional
from dataclasses import dataclass, field

from ..reporting.blocks import BaseBlock, Finding


class EventType(Enum):
    KUBERNETES_TOPOLOGY_CHANGE = 1
    PROMETHEUS = 2
    MANUAL_TRIGGER = 3
    SCHEDULED_TRIGGER = 4


# Right now:
# 1. this is a dataclass but we need to make all fields optional in subclasses because of https://stackoverflow.com/questions/51575931/
# 2. this can't be a pydantic BaseModel because of various pydantic bugs (see https://github.com/samuelcolvin/pydantic/pull/2557)
# once the pydantic PR that addresses those issues is merged, this should be a pydantic class
@dataclass
class BaseEvent:
    # TODO: just like you can add generic reporting blocks, should we allow attaching persistent context too?
    report_blocks: List[BaseBlock] = field(default_factory=list)
    # some chat APIs allow attachment blocks which are formatted differently
    report_attachment_blocks: List[BaseBlock] = field(default_factory=list)
    report_title: str = ""
    prometheus_url: Optional[str] = None

    finding: Optional[Finding] = None

    named_sinks: Optional[List[str]] = None



from functools import wraps

# from https://stackoverflow.com/a/14412901/495995
def doublewrap(f):
    """
    a decorator decorator, allowing the decorator to be used as:
    @decorator(with, arguments, and=kwargs)
    or
    @decorator
    """

    @wraps(f)
    def new_dec(*args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
            # actual decorated function
            return f(args[0])
        else:
            # decorator arguments
            return lambda realf: f(realf, *args, **kwargs)

    return new_dec



import logging
from threading import Thread
from queue import Queue


class TaskQueue(Queue):
    def __init__(self, num_workers=1):
        Queue.__init__(self)
        logging.info(f"Initialized task queue: {num_workers} workers")
        self.num_workers = num_workers
        self.start_workers()

    def add_task(self, task, *args, **kwargs):
        args = args or ()
        kwargs = kwargs or {}
        self.put((task, args, kwargs))

    def start_workers(self):
        for i in range(self.num_workers):
            t = Thread(target=self.worker)
            t.daemon = True
            t.start()

    def worker(self):
        while True:
            item, args, kwargs = self.get()
            item(*args, **kwargs)
            self.task_done()



from pydantic import BaseModel
from typing import List

from ..model.k8s_operation_type import K8sOperationType


class TriggerParams(BaseModel):
    trigger_name: str = None
    alert_name: str = None
    pod_name_prefix: str = None
    instance_name_prefix: str = None
    name_prefix: str = None
    namespace_prefix: str = None
    status: str = None
    kind: str = None
    operation: K8sOperationType = None
    repeat: int = None
    seconds_delay: int = None
    delays: List[int] = None



from pydantic.main import BaseModel

from ..sink_config import SinkConfigBase
from ....integrations.slack import SlackSender
from ...reporting.blocks import Finding
from ..sink_base import SinkBase


class SlackSinkConfig(BaseModel):
    slack_channel: str
    api_key: str


class SlackSink(SinkBase):
    def __init__(self, sink_config: SinkConfigBase):
        super().__init__(sink_config)
        config = SlackSinkConfig(**sink_config.params)
        self.slack_channel = config.slack_channel
        self.slack_sender = SlackSender(config.api_key)

    def write_finding(self, finding: Finding):
        self.slack_sender.send_finding_to_slack(
            finding, self.slack_channel, self.sink_name
        )



import uuid
from datetime import datetime

import flask
from ...core.model.cloud_event import CloudEvent
from ...core.model.events import EventType


def parse_incoming_manual_trigger(request: flask.Request) -> CloudEvent:
    trigger_name = request.form.get("trigger_name")
    if trigger_name is None:
        raise Exception(f"manual trigger is missing trigger_name. request={request}")

    event_data = request.form.to_dict()
    event_data["description"] = f"manual trigger for playbook {trigger_name}"

    return CloudEvent(
        specversion="1.0",
        type=EventType.MANUAL_TRIGGER.name,
        source=EventType.MANUAL_TRIGGER.name,
        subject=trigger_name,
        id=str(uuid.uuid4()),
        time=datetime.now(),
        datacontenttype="application/json",
        data=event_data,
    )



import logging
from functools import wraps
from dataclasses import dataclass, field
import pydantic
from typing import List

from ..base_handler import handle_event
from ...core.active_playbooks import activate_playbook, register_playbook
from ...core.model.playbook_hash import playbook_hash
from ...utils.decorators import doublewrap
from ...core.model.trigger_params import TriggerParams
from ...core.model.cloud_event import CloudEvent
from ...core.model.events import EventType, BaseEvent


@dataclass
class ManualTriggerEvent(BaseEvent):
    trigger_name: str = ""
    data: dict = field(default_factory=dict)


@doublewrap
def on_manual_trigger(func):
    return register_playbook(
        func, deploy_manual_trigger, TriggerParams(trigger_name=func.__name__)
    )


def deploy_manual_trigger(
    func, trigger_params: TriggerParams, named_sinks: List[str], action_params=None
):
    @wraps(func)
    def wrapper(cloud_event: CloudEvent):
        trigger_event = ManualTriggerEvent(
            trigger_name=cloud_event.subject, data=cloud_event.data
        )
        logging.debug(
            f"checking if we should run manually triggered playbook {func}. trigger_name in request is {trigger_event.trigger_name} and playbook trigger_name is {trigger_params.trigger_name}"
        )

        if trigger_event.trigger_name != trigger_params.trigger_name:
            logging.debug("not running")
            return

        try:
            return handle_event(
                func, trigger_event, action_params, "manual", named_sinks
            )
        except pydantic.error_wrappers.ValidationError as e:
            logging.error(f"{e}")

    playbook_id = playbook_hash(func, trigger_params, action_params)
    activate_playbook(EventType.MANUAL_TRIGGER, wrapper, func, playbook_id)
    return wrapper



from pydantic import BaseModel
from typing import Optional, List

from .trigger_params import TriggerParams


class PlaybookDeployConfig(BaseModel):
    name: str = None
    sinks: List[str] = None
    trigger_params: Optional[TriggerParams] = TriggerParams()
    action_params: Optional[dict] = {}



# This file was autogenerated. Do not edit.

from dataclasses import dataclass
from typing import Union
from ..base_event import K8sBaseEvent
from ..custom_models import RobustaPod, RobustaDeployment
from hikaru.model.rel_1_16.v1 import Pod as v1Pod
from hikaru.model.rel_1_16.v1 import ReplicaSet as v1ReplicaSet
from hikaru.model.rel_1_16.v1 import DaemonSet as v1DaemonSet
from hikaru.model.rel_1_16.v1 import Deployment as v1Deployment
from hikaru.model.rel_1_16.v1 import StatefulSet as v1StatefulSet
from hikaru.model.rel_1_16.v1 import Service as v1Service
from hikaru.model.rel_1_16.v1 import ConfigMap as v1ConfigMap
from hikaru.model.rel_1_16.v1 import Event as v1Event
from hikaru.model.rel_1_16.v1 import (
    HorizontalPodAutoscaler as v1HorizontalPodAutoscaler,
)
from hikaru.model.rel_1_16.v1 import Node as v1Node
from hikaru.model.rel_1_16.v2beta1 import Pod as v2beta1Pod
from hikaru.model.rel_1_16.v2beta1 import ReplicaSet as v2beta1ReplicaSet
from hikaru.model.rel_1_16.v2beta1 import DaemonSet as v2beta1DaemonSet
from hikaru.model.rel_1_16.v2beta1 import Deployment as v2beta1Deployment
from hikaru.model.rel_1_16.v2beta1 import StatefulSet as v2beta1StatefulSet
from hikaru.model.rel_1_16.v2beta1 import Service as v2beta1Service
from hikaru.model.rel_1_16.v2beta1 import ConfigMap as v2beta1ConfigMap
from hikaru.model.rel_1_16.v2beta1 import Event as v2beta1Event
from hikaru.model.rel_1_16.v2beta1 import (
    HorizontalPodAutoscaler as v2beta1HorizontalPodAutoscaler,
)
from hikaru.model.rel_1_16.v2beta1 import Node as v2beta1Node
from hikaru.model.rel_1_16.v2beta2 import Pod as v2beta2Pod
from hikaru.model.rel_1_16.v2beta2 import ReplicaSet as v2beta2ReplicaSet
from hikaru.model.rel_1_16.v2beta2 import DaemonSet as v2beta2DaemonSet
from hikaru.model.rel_1_16.v2beta2 import Deployment as v2beta2Deployment
from hikaru.model.rel_1_16.v2beta2 import StatefulSet as v2beta2StatefulSet
from hikaru.model.rel_1_16.v2beta2 import Service as v2beta2Service
from hikaru.model.rel_1_16.v2beta2 import ConfigMap as v2beta2ConfigMap
from hikaru.model.rel_1_16.v2beta2 import Event as v2beta2Event
from hikaru.model.rel_1_16.v2beta2 import (
    HorizontalPodAutoscaler as v2beta2HorizontalPodAutoscaler,
)
from hikaru.model.rel_1_16.v2beta2 import Node as v2beta2Node


@dataclass
class PodEvent(K8sBaseEvent):
    obj: RobustaPod
    old_obj: RobustaPod


@dataclass
class ReplicaSetEvent(K8sBaseEvent):
    obj: Union[v1ReplicaSet, v2beta1ReplicaSet, v2beta2ReplicaSet]
    old_obj: Union[v1ReplicaSet, v2beta1ReplicaSet, v2beta2ReplicaSet]


@dataclass
class DaemonSetEvent(K8sBaseEvent):
    obj: Union[v1DaemonSet, v2beta1DaemonSet, v2beta2DaemonSet]
    old_obj: Union[v1DaemonSet, v2beta1DaemonSet, v2beta2DaemonSet]


@dataclass
class DeploymentEvent(K8sBaseEvent):
    obj: RobustaDeployment
    old_obj: RobustaDeployment


@dataclass
class StatefulSetEvent(K8sBaseEvent):
    obj: Union[v1StatefulSet, v2beta1StatefulSet, v2beta2StatefulSet]
    old_obj: Union[v1StatefulSet, v2beta1StatefulSet, v2beta2StatefulSet]


@dataclass
class ServiceEvent(K8sBaseEvent):
    obj: Union[v1Service, v2beta1Service, v2beta2Service]
    old_obj: Union[v1Service, v2beta1Service, v2beta2Service]


@dataclass
class ConfigMapEvent(K8sBaseEvent):
    obj: Union[v1ConfigMap, v2beta1ConfigMap, v2beta2ConfigMap]
    old_obj: Union[v1ConfigMap, v2beta1ConfigMap, v2beta2ConfigMap]


@dataclass
class EventEvent(K8sBaseEvent):
    obj: Union[v1Event, v2beta1Event, v2beta2Event]
    old_obj: Union[v1Event, v2beta1Event, v2beta2Event]


@dataclass
class HorizontalPodAutoscalerEvent(K8sBaseEvent):
    obj: Union[
        v1HorizontalPodAutoscaler,
        v2beta1HorizontalPodAutoscaler,
        v2beta2HorizontalPodAutoscaler,
    ]
    old_obj: Union[
        v1HorizontalPodAutoscaler,
        v2beta1HorizontalPodAutoscaler,
        v2beta2HorizontalPodAutoscaler,
    ]


@dataclass
class NodeEvent(K8sBaseEvent):
    obj: Union[v1Node, v2beta1Node, v2beta2Node]
    old_obj: Union[v1Node, v2beta1Node, v2beta2Node]


@dataclass
class KubernetesAnyEvent(K8sBaseEvent):
    obj: Union[
        v2beta1ConfigMap,
        RobustaDeployment,
        v1HorizontalPodAutoscaler,
        v1ConfigMap,
        v2beta2StatefulSet,
        v1Node,
        v1Event,
        v1ReplicaSet,
        v2beta1Event,
        v2beta2HorizontalPodAutoscaler,
        v2beta2ConfigMap,
        v1DaemonSet,
        v1StatefulSet,
        v2beta2DaemonSet,
        v2beta1ReplicaSet,
        v2beta2Node,
        v2beta2ReplicaSet,
        v2beta2Event,
        RobustaPod,
        v2beta1StatefulSet,
        v2beta1HorizontalPodAutoscaler,
        v1Service,
        v2beta1Node,
        v2beta2Service,
        v2beta1DaemonSet,
        v2beta1Service,
    ]
    old_obj: Union[
        v2beta1ConfigMap,
        RobustaDeployment,
        v1HorizontalPodAutoscaler,
        v1ConfigMap,
        v2beta2StatefulSet,
        v1Node,
        v1Event,
        v1ReplicaSet,
        v2beta1Event,
        v2beta2HorizontalPodAutoscaler,
        v2beta2ConfigMap,
        v1DaemonSet,
        v1StatefulSet,
        v2beta2DaemonSet,
        v2beta1ReplicaSet,
        v2beta2Node,
        v2beta2ReplicaSet,
        v2beta2Event,
        RobustaPod,
        v2beta1StatefulSet,
        v2beta1HorizontalPodAutoscaler,
        v1Service,
        v2beta1Node,
        v2beta2Service,
        v2beta1DaemonSet,
        v2beta1Service,
    ]


KIND_TO_EVENT_CLASS = {
    "Pod": PodEvent,
    "ReplicaSet": ReplicaSetEvent,
    "DaemonSet": DaemonSetEvent,
    "Deployment": DeploymentEvent,
    "StatefulSet": StatefulSetEvent,
    "Service": ServiceEvent,
    "ConfigMap": ConfigMapEvent,
    "Event": EventEvent,
    "HorizontalPodAutoscaler": HorizontalPodAutoscalerEvent,
    "Node": NodeEvent,
}



import regex
from hikaru import HikaruBase


def update_item_attr(obj: HikaruBase, attr_key: str, attr_value):
    path_parts = regex.split("\\[|\\].|\\]|\\.", attr_key)
    parent_item = obj.object_at_path(path_parts[0 : len(path_parts) - 1])
    last_part = path_parts[len(path_parts) - 1]
    if type(parent_item) == dict:
        parent_item[last_part] = attr_value
    elif type(parent_item) == list:
        parent_item[int(last_part)] = attr_value
    else:
        setattr(parent_item, last_part, attr_value)



import base64
import json
import logging
import time
import traceback

from hikaru.model import Deployment, StatefulSetList, DaemonSetList, ReplicaSetList
from typing import List, Dict
from pydantic import BaseModel
import threading

from ..sink_config import SinkConfigBase
from ...model.env_vars import DISCOVERY_PERIOD_SEC
from ...model.services import ServiceInfo
from ...reporting.blocks import Finding
from .dal.supabase_dal import SupabaseDal
from ..sink_base import SinkBase
from ...discovery.top_service_resolver import TopServiceResolver


class RobustaSinkConfig(BaseModel):
    token: str


class RobustaToken(BaseModel):
    store_url: str
    api_key: str
    account_id: str
    email: str
    password: str


class RobustaSink(SinkBase):
    def __init__(
        self,
        sink_config: SinkConfigBase,
        cluster_name: str,
    ):
        super().__init__(sink_config)
        config = RobustaSinkConfig(**sink_config.params)
        robusta_token = RobustaToken(**json.loads(base64.b64decode(config.token)))
        self.dal = SupabaseDal(
            robusta_token.store_url,
            robusta_token.api_key,
            robusta_token.account_id,
            robusta_token.email,
            robusta_token.password,
            sink_config.sink_name,
            cluster_name,
        )
        # start service discovery
        self.__active = True
        self.__discovery_period_sec = DISCOVERY_PERIOD_SEC
        self.__services_cache: Dict[str, ServiceInfo] = {}
        self.__thread = threading.Thread(target=self.__discover_services)
        self.__thread.start()

    def stop(self):
        self.__active = False

    def write_finding(self, finding: Finding):
        self.dal.persist_finding(finding)

    # service discovery impl
    def __publish_service(self, serviceInfo: ServiceInfo):
        logging.debug(f"publishing to {self.sink_name} service {serviceInfo} ")
        self.dal.persist_service(serviceInfo)

    def __is_cached(self, service_info: ServiceInfo):
        cache_key = service_info.get_service_key()
        return self.__services_cache.get(cache_key) is not None

    def __publish_new_services(self, active_services: List):
        active_services_keys = set()
        for service in active_services:
            service_info = ServiceInfo(
                name=service.metadata.name,
                namespace=service.metadata.namespace,
                service_type=service.kind,
            )
            cache_key = service_info.get_service_key()
            active_services_keys.add(cache_key)
            cached_service = self.__services_cache.get(cache_key)
            if not cached_service or cached_service != service_info:
                self.__publish_service(service_info)
                self.__services_cache[cache_key] = service_info

        # delete cached services that aren't active anymore
        cache_keys = list(self.__services_cache.keys())
        for service_key in cache_keys:
            if service_key not in active_services_keys:
                del self.__services_cache[service_key]

        # handle delete services
        persisted_services = self.dal.get_active_services()
        deleted_services = [
            service for service in persisted_services if not self.__is_cached(service)
        ]
        for deleted_service in deleted_services:
            deleted_service.deleted = True
            self.__publish_service(deleted_service)

        # save the cached services in the resolver.
        TopServiceResolver.store_cached_services(list(self.__services_cache.values()))

    def __discover_services(self):
        while self.__active:
            try:
                current_services = Deployment.listDeploymentForAllNamespaces().obj.items
                current_services.extend(
                    StatefulSetList.listStatefulSetForAllNamespaces().obj.items
                )
                current_services.extend(
                    DaemonSetList.listDaemonSetForAllNamespaces().obj.items
                )
                current_services.extend(
                    [
                        rs
                        for rs in ReplicaSetList.listReplicaSetForAllNamespaces().obj.items
                        if rs.metadata.ownerReferences is None
                    ]
                )
                self.__publish_new_services(current_services)
            except Exception as e:
                logging.error(
                    f"Failed to run periodic service discovery for {self.sink_name}",
                    traceback.print_exc(),
                )

            time.sleep(self.__discovery_period_sec)

        logging.info(f"Service discovery for sink {self.sink_name} ended.")



import logging
from typing import Dict, List, Optional

from .datadog.datadog_sink import DataDogSink
from .kafka.kafka_sink import KafkaSink
from .robusta.robusta_sink import RobustaSink
from .slack.slack_sink import SlackSink
from .sink_config import SinkConfigBase
from .sink_base import SinkBase
from ..reporting.consts import SinkType


class SinkManager:
    sinks: Dict[str, SinkBase] = {}
    cluster_name: str

    @staticmethod
    def get_sink_by_name(sink_name: str) -> Optional[SinkBase]:
        return SinkManager.sinks.get(sink_name)

    @staticmethod
    def __add_sink(sink_config):
        try:
            if sink_config.sink_type == SinkType.KAFKA.value:
                SinkManager.sinks[sink_config.sink_name] = KafkaSink(sink_config)
            elif sink_config.sink_type == SinkType.ROBUSTA.value:
                SinkManager.sinks[sink_config.sink_name] = RobustaSink(
                    sink_config, SinkManager.cluster_name
                )
            elif sink_config.sink_type == SinkType.DATADOG.value:
                SinkManager.sinks[sink_config.sink_name] = DataDogSink(
                    sink_config, SinkManager.cluster_name
                )
            elif sink_config.sink_type == SinkType.SLACK.value:
                SinkManager.sinks[sink_config.sink_name] = SlackSink(sink_config)
            else:
                raise Exception(f"Sink not supported {sink_config.sink_type}")
        except Exception as e:
            logging.error(
                f"Error configuring sink {sink_config.sink_name} of type {sink_config.sink_type}: {e}"
            )

    @staticmethod
    def update_sinks_config(sinks_config: List[SinkConfigBase], cluster_name: str):
        SinkManager.cluster_name = cluster_name
        new_sink_names = [sink_config.sink_name for sink_config in sinks_config]

        # remove deleted sinks
        deleted_sink_names = [
            sink_name
            for sink_name in SinkManager.sinks.keys()
            if sink_name not in new_sink_names
        ]
        for deleted_sink in deleted_sink_names:
            logging.info(f"Deleting sink {deleted_sink}")
            SinkManager.sinks[deleted_sink].stop()
            del SinkManager.sinks[deleted_sink]

        # create new sinks, or update existing if changed
        for sink_config in sinks_config:
            # temporary workaround to skip the default and unconfigured robusta token
            if (
                sink_config.sink_type == "robusta"
                and sink_config.params.get("token") == "<ROBUSTA_ACCOUNT_TOKEN>"
            ):
                continue

            if sink_config.sink_name not in SinkManager.sinks.keys():
                logging.info(
                    f"Adding {sink_config.sink_type} sink named {sink_config.sink_name}"
                )
                SinkManager.__add_sink(sink_config)
            elif sink_config.params != SinkManager.sinks[sink_config.sink_name].params:
                logging.info(
                    f"Updating {sink_config.sink_type} sink named {sink_config.sink_name}"
                )
                SinkManager.sinks[sink_config.sink_name].stop()
                SinkManager.__add_sink(sink_config)



from pydantic import BaseModel

# TODO: this should not be part of the Robusta API
# the only usage from a playbook is in event_publisher.py
from ..core.active_playbooks import run_playbooks
from ..core.model.base_params import *
from ..core.sinks import *
from ..core.sinks.sink_base import *
from ..core.sinks.sink_config import *
from ..core.sinks.sink_manager import *
from ..core.sinks.kafka import *
from ..core.reporting.consts import *
from ..core.reporting.callbacks import *
from ..core.reporting.custom_rendering import *
from ..core.schedule.model import *
from ..core.model.env_vars import *
from ..core.model.k8s_operation_type import *

from ..core.reporting import *
from ..integrations.kubernetes.custom_models import *
from ..integrations.kubernetes.autogenerated.triggers import *
from ..integrations.kubernetes.autogenerated.events import *
from ..integrations.kubernetes.autogenerated.models import *
from ..integrations.prometheus.triggers import *
from ..integrations.prometheus.incoming_handler import prometheus_cloud_event
from ..integrations.prometheus.models import *
from ..integrations.prometheus.utils import *
from ..integrations.slack.sender import *
from ..integrations.grafana import *
from ..integrations.manual.triggers import *
from ..integrations.scheduled.triggers import *
from ..integrations.git.git_repo import *
from ..integrations.resource_analysis.kubernetes_node_analyzer import NodeAnalyzer
from ..integrations.prometheus.utils import PrometheusDiscovery
from ..core.persistency.in_memory import get_persistent_data
from ..utils.rate_limiter import RateLimiter
from ..utils.common import is_matching_diff
from ..runner.object_updater import *



from datetime import datetime
import logging
from typing import List

from grafana_api.grafana_face import GrafanaFace

from ..utils.service_discovery import find_service_url


class Grafana:
    def __init__(self, api_key, grafana_url=None):
        """
        Create a new connection to Grafana.
        :param grafana_url: if None, then attempt to discover the address of an in-cluster Grafana service
        """
        if grafana_url is None:
            grafana_url = find_service_url("app.kubernetes.io/name=grafana")
        protocol_host = grafana_url.split("://")
        logging.debug(
            f"Grafana params: protocol - {protocol_host[0]} host - {protocol_host[1]}"
        )
        self.grafana = GrafanaFace(
            auth=api_key, protocol=protocol_host[0], host=protocol_host[1]
        )

    def add_line_to_dashboard(
        self,
        dashboard_uid: str,
        text: str,
        time: datetime = None,
        tags: List[str] = [],
        panel_substring: str = None,
    ):
        if time is None:
            time = datetime.now()
        self.__add_annotation(
            dashboard_uid,
            text,
            start_time=time,
            tags=tags,
            panel_substring=panel_substring,
        )

    def add_range_to_dashboard(
        self,
        dashboard_uid: str,
        text: str,
        start_time: datetime,
        end_time: datetime,
        tags: List[str] = [],
        panel_substring: str = None,
    ):
        self.__add_annotation(
            dashboard_uid,
            text,
            start_time=start_time,
            end_time=end_time,
            tags=tags,
            panel_substring=panel_substring,
        )

    def __add_annotation(
        self,
        dashboard_uid,
        text,
        start_time,
        end_time=None,
        tags=[],
        panel_substring=None,
    ):
        dashboard = self.grafana.dashboard.get_dashboard(dashboard_uid)["dashboard"]
        dashboard_id = dashboard["id"]

        # grafana wants the timestamp as an int with millisecond resolution
        start_time = int(start_time.timestamp()) * 1000
        if end_time is not None:
            end_time = int(end_time.timestamp()) * 1000

        # add an annotation for the entire dashboard
        if panel_substring is None:
            resp = self.grafana.annotations.add_annotation(
                dashboard_id=dashboard_id,
                text=text,
                tags=tags,
                time_from=start_time,
                time_to=end_time,
            )
            logging.debug(f"grafana dashboard annotation response {resp}")
        # add an annotation to specific panels only
        else:
            panel_ids = self.__get_panels_with_subtring(dashboard, panel_substring)
            for panel_id in panel_ids:
                resp = self.grafana.annotations.add_annotation(
                    dashboard_id=dashboard_id,
                    panel_id=panel_id,
                    text=text,
                    tags=tags,
                    time_from=start_time,
                    time_to=end_time,
                )
                logging.debug(f"grafana panel annotation response {resp}")

    def __get_panels_with_subtring(self, dashboard, panel_substring):
        panel_ids = []
        for row in dashboard["rows"]:
            for panel in row["panels"]:
                if panel_substring.lower() in panel["title"].lower():
                    panel_ids.append(panel["id"])
        return panel_ids



import hashlib

from ...utils.function_hashes import get_function_hash
from ...core.model.trigger_params import TriggerParams


def playbook_hash(func, trigger_params: TriggerParams, action_params):
    hash_input = (
        f"{get_function_hash(func)}"
        + ("None" if trigger_params is None else trigger_params.json())
        + ("None" if action_params is None else action_params.json())
    )
    return hashlib.md5(hash_input.encode()).hexdigest()



from typing import List

from .blocks import *


def convert_svg_to_png(svg: bytes) -> bytes:
    # we import cairosvg here and not globally because in some environments it isn't trivially installed (e.g. windows)
    # and we don't want to throw an exception globally when it isn't around
    import cairosvg

    return cairosvg.svg2png(bytestring=svg)


def add_pngs_for_all_svgs(blocks: List[FileBlock]):
    new_blocks = blocks.copy()
    for b in blocks:
        if not isinstance(b, FileBlock):
            continue
        if not b.filename.endswith(".svg"):
            continue
        conversion = convert_svg_to_png(b.contents)
        new_blocks.append(FileBlock(b.filename.replace(".svg", ".png"), conversion))
    return new_blocks



import json
import logging
from typing import List, Optional
from threading import Lock

import kubernetes
from hikaru.model import ObjectMeta

from ..model.env_vars import INSTALLATION_NAMESPACE
from ...core.schedule.model import ScheduledJob
from ...integrations.kubernetes.autogenerated.v1.models import ConfigMap

JOBS_CONFIGMAP_NAME = "scheduled-jobs"
CONFIGMAP_NAMESPACE = INSTALLATION_NAMESPACE


class SchedulerDal:
    mutex = Lock()

    def __init__(self):
        self.__init_scheduler_dal()

    def __load_config_map(self) -> ConfigMap:
        return ConfigMap.readNamespacedConfigMap(
            JOBS_CONFIGMAP_NAME, CONFIGMAP_NAMESPACE
        ).obj

    def __init_scheduler_dal(self):
        try:
            self.__load_config_map()
        except kubernetes.client.exceptions.ApiException as e:
            # we only want to catch exceptions because the config map doesn't exist
            if e.reason != "Not Found":
                raise
            # job states configmap doesn't exists, create it
            self.mutex.acquire()
            try:
                conf_map = ConfigMap(
                    metadata=ObjectMeta(
                        name=JOBS_CONFIGMAP_NAME, namespace=CONFIGMAP_NAMESPACE
                    )
                )
                conf_map.createNamespacedConfigMap(conf_map.metadata.namespace)
                logging.info(
                    f"created jobs states configmap {JOBS_CONFIGMAP_NAME} {CONFIGMAP_NAMESPACE}"
                )
            finally:
                self.mutex.release()

    def save_scheduled_job(self, job: ScheduledJob):
        self.mutex.acquire()
        try:
            confMap = self.__load_config_map()
            confMap.data[job.job_id] = job.json()
            confMap.replaceNamespacedConfigMap(
                confMap.metadata.name, confMap.metadata.namespace
            )
        finally:
            self.mutex.release()

    def get_scheduled_job(self, job_id: str) -> Optional[ScheduledJob]:
        state_data = self.__load_config_map().data.get(job_id)
        return (
            ScheduledJob(**json.loads(state_data)) if state_data is not None else None
        )

    def del_scheduled_job(self, job_id: str):
        self.mutex.acquire()
        try:
            confMap = self.__load_config_map()
            if confMap.data.get(job_id) is not None:
                del confMap.data[job_id]
                confMap.replaceNamespacedConfigMap(
                    confMap.metadata.name, confMap.metadata.namespace
                )
        finally:
            self.mutex.release()

    def list_scheduled_jobs(self) -> List[ScheduledJob]:
        return [
            self.get_scheduled_job(job_id)
            for job_id in self.__load_config_map().data.keys()
        ]



import os
import subprocess
import time
import traceback
import click_spinner
from typing import List, Optional

import yaml

import typer

from ..cli.utils import (
    log_title,
    fetch_runner_logs,
    exec_in_robusta_runner,
    PLAYBOOKS_DIR,
)


app = typer.Typer()


@app.command()
def push(
    playbooks_directory: str = typer.Argument(
        ...,
        help="Local playbooks code directory",
    ),
    namespace: str = typer.Option(
        "robusta",
        help="Robusta installation namespace",
    ),
):
    """Load custom playbooks code"""
    log_title("Uploading playbooks code...")
    with fetch_runner_logs(namespace):
        subprocess.check_call(
            f"kubectl create configmap -n {namespace} robusta-custom-playbooks --from-file {playbooks_directory} -o yaml --dry-run | kubectl apply -f -",
            shell=True,
        )
        subprocess.check_call(
            f'kubectl annotate pods -n {namespace} --all --overwrite "playbooks-last-modified={time.time()}"',
            shell=True,
        )
        time.sleep(
            5
        )  # wait five seconds for the runner to actually reload the playbooks
    log_title("Loaded custom playbooks code!")


@app.command()
def configure(
    config_file: str = typer.Argument(
        ...,
        help="Playbooks configuration file",
    ),
    namespace: str = typer.Option(
        "robusta",
        help="Robusta installation namespace",
    ),
):
    """deploy playbooks configuration"""
    log_title("Configuring playbooks...")
    with fetch_runner_logs(namespace):
        subprocess.check_call(
            f"kubectl create configmap -n {namespace} robusta-playbooks-config --from-file active_playbooks.yaml={config_file} -o yaml --dry-run | kubectl apply -f -",
            shell=True,
        )
        subprocess.check_call(
            f'kubectl annotate pods -n {namespace} --all --overwrite "playbooks-last-modified={time.time()}"',
            shell=True,
        )
        time.sleep(
            5
        )  # wait five seconds for the runner to actually reload the playbooks
    log_title("Deployed playbooks!")


def get_playbooks_config(namespace: str):
    configmap_content = subprocess.check_output(
        f"kubectl get configmap -n {namespace} robusta-playbooks-config -o yaml",
        shell=True,
    )
    return yaml.safe_load(configmap_content)


def get_custom_playbooks(namespace: str):
    configmap_content = subprocess.check_output(
        f"kubectl get configmap -n {namespace} robusta-custom-playbooks -o yaml",
        shell=True,
    )
    return yaml.safe_load(configmap_content)


@app.command()
def pull(
    playbooks_directory: str = typer.Argument(
        ...,
        help="Local target directory",
    ),
    namespace: str = typer.Option(
        "robusta",
        help="Installation namespace",
    ),
):
    """pull cluster deployed playbooks"""
    if not playbooks_directory:
        playbooks_directory = os.path.join(os.getcwd(), PLAYBOOKS_DIR)

    log_title(f"Pulling playbooks into {playbooks_directory} ")

    try:
        playbooks_config = get_custom_playbooks(namespace)

        for file_name in playbooks_config["data"].keys():
            playbook_file = os.path.join(playbooks_directory, file_name)
            with open(playbook_file, "w") as f:
                f.write(playbooks_config["data"][file_name])

    except Exception as e:
        typer.echo(f"Failed to pull deployed playbooks {e}", traceback.print_exc())


def print_yaml_if_not_none(key: str, json_dict: dict):
    if json_dict.get(key):
        json = {}
        json[key] = json_dict.get(key)
        typer.echo(f"{yaml.dump(json)}")


@app.command("list")
def list_(
    namespace: str = typer.Option(
        "robusta",
        help="Installation namespace",
    ),
):  # not named list as that would shadow the builtin list function
    """list current active playbooks"""
    typer.echo(f"Getting deployed playbooks list...")
    with click_spinner.spinner():
        playbooks_config = get_playbooks_config(namespace)

    active_playbooks_file = playbooks_config["data"]["active_playbooks.yaml"]
    active_playbooks_yaml = yaml.safe_load(active_playbooks_file)
    for playbook in active_playbooks_yaml["active_playbooks"]:
        log_title(f"Playbook: {playbook['name']}")
        print_yaml_if_not_none("name", playbook)
        print_yaml_if_not_none("sinks", playbook)
        print_yaml_if_not_none("trigger_params", playbook)
        print_yaml_if_not_none("action_params", playbook)


@app.command()
def show_config(
    namespace: str = typer.Option(
        "robusta",
        help="Installation namespace",
    ),
):
    """fetch and show active_playbooks.yaml from cluster"""
    typer.echo("connecting to cluster...")
    with click_spinner.spinner():
        playbooks_config = get_playbooks_config(namespace)
    active_playbooks_file = playbooks_config["data"]["active_playbooks.yaml"]
    log_title("Contents of active_playbooks.yaml:")
    typer.echo(active_playbooks_file)


@app.command()
def trigger(
    trigger_name: str,
    param: Optional[List[str]] = typer.Argument(
        None,
        help="data to send to playbook (can be used multiple times)",
        metavar="key=value",
    ),
    namespace: str = typer.Option(
        "robusta",
        help="Install Robusta on the specified custom namespace",
    ),
):
    """trigger a manually run playbook"""
    log_title("Triggering playbook...")
    trigger_params = " ".join([f"-F '{p}'" for p in param])
    with fetch_runner_logs(namespace=namespace):
        cmd = f"curl -X POST -F 'trigger_name={trigger_name}' {trigger_params} http://localhost:5000/api/trigger"
        exec_in_robusta_runner(
            cmd,
            namespace=namespace,
            tries=3,
            error_msg="Cannot trigger playbook - usually this means Robusta just started. Will try again",
        )
        typer.echo("\n")
    log_title("Done!")


if __name__ == "__main__":
    app()



from pydantic import BaseModel
from typing import Dict, Any, List


class GenParams(BaseModel):
    name: str
    params: Dict[Any, Any] = {}


class SlackParams(BaseModel):
    slack_channel: str


class NodeNameParams(BaseModel):
    node_name: str = None


class PodParams(BaseModel):
    pod_name: str = None
    pod_namespace: str = None


class NamespacedKubernetesObjectParams(BaseModel):
    name: str = None
    namespace: str = None



import datetime
import uuid

import flask

from ...core.model.cloud_event import CloudEvent
from ...core.model.events import EventType


def parse_incoming_prometheus_alerts(request: flask.Request) -> CloudEvent:
    event_data = request.get_json()
    event_data["description"] = "prometheus alerts group"
    return prometheus_cloud_event(event_data)


def prometheus_cloud_event(event_data):
    cloud_event = CloudEvent(
        **{
            "specversion": "1.0",
            "type": EventType.PROMETHEUS.name,
            "source": EventType.PROMETHEUS.name,
            "subject": "",
            "id": str(uuid.uuid4()),
            "datacontenttype": "application/json",
            "time": datetime.datetime.now(),
            "data": event_data,
        }
    )
    return cloud_event



import datetime
import logging
import os
import re
import time
import traceback
from typing import List, Optional

from kubernetes import config
from kubernetes.client.api import core_v1_api
from kubernetes.client.rest import ApiException
from kubernetes.stream import stream

from hikaru.model import Job

RUNNING_STATE = "Running"

try:
    if os.getenv("KUBERNETES_SERVICE_HOST"):
        config.load_incluster_config()
    else:
        config.load_kube_config()
except config.config_exception.ConfigException as e:
    logging.warning(f"Running without kube-config! e={e}")


core_v1 = core_v1_api.CoreV1Api()

default_exec_command = ["/bin/sh", "-c"]


def wait_until(
    read_function, predicate_function, timeout_sec: float, backoff_wait_sec: float
):
    """
    repeatedly calls predicate_function(read_function)) until predicate_function returns True or we timeout
    return the last result of read_function() on success and raises an exception on timeout
    between attempts, we wait backoff_wait_sec seconds
    """
    start_time_sec = time.time()

    while start_time_sec + timeout_sec > time.time():
        try:
            resp = read_function()
            if predicate_function(resp):
                return resp
        except ApiException as e:
            logging.error(f"failed calling read_function {traceback.format_exc()}")

        time.sleep(backoff_wait_sec)

    raise Exception("Failed to reach wait condition")


def wait_until_job_complete(job: Job, timeout):
    """
    wait until a kubernetes Job object either succeeds or fails at least once
    """

    def is_job_complete(j: Job) -> bool:
        return j.status.completionTime is not None or j.status.failed is not None

    return wait_until(
        lambda: Job.readNamespacedJob(job.metadata.name, job.metadata.namespace).obj,
        is_job_complete,
        timeout,
        5,
    )


# TODO: refactor to use wait_until function
def wait_for_pod_status(
    name, namespace, status: str, timeout_sec: float, backoff_wait_sec: float
) -> str:
    pod_details = f"pod status: {name} {namespace} {status} {timeout_sec}"
    logging.debug(f"waiting for {pod_details}")

    start_time_sec = time.time()
    while start_time_sec + timeout_sec > time.time():
        try:
            resp = core_v1.read_namespaced_pod_status(name, namespace)

            if resp.status.phase == status:
                logging.debug(f"reached {pod_details}")
                return status

        except ApiException as e:
            logging.error(
                f"failed to get pod status {name} {namespace} {traceback.format_exc()}"
            )

        time.sleep(backoff_wait_sec)

    logging.debug(f"failed to reach {pod_details}")
    return "FAIL"


def exec_shell_command(name, shell_command: str, namespace="default", container=""):
    commands = default_exec_command.copy()
    commands.append(shell_command)
    return exec_commands(name, commands, namespace, container)


def get_pod_logs(
    name,
    namespace="default",
    container="",
    previous=None,
    tail_lines=None,
    since_seconds=None,
):
    resp = None
    try:
        resp = core_v1.read_namespaced_pod_log(
            name,
            namespace,
            container=container,
            previous=previous,
            tail_lines=tail_lines,
            since_seconds=since_seconds,
        )

    except ApiException as e:
        if e.status != 404:
            logging.exception(f"failed to get pod logs {name} {namespace} {container}")
            resp = ""

    logging.debug(f"get logs {resp}")
    return resp


def prepare_pod_command(cmd) -> Optional[List[str]]:
    if type(cmd) == list:
        return cmd
    elif type(cmd) == str:
        return cmd.split(" ")  # cmd need to be a list of strings
    elif cmd is None:
        return None
    else:
        logging.exception(f"cmd {cmd} has unknown type {type(cmd)}")
        return cmd


def exec_commands(name, exec_command, namespace="default", container=""):
    logging.debug(
        f"Executing command name: {name} command: {exec_command} namespace: {namespace} container: {container}"
    )
    resp = None

    # verify pod state before connecting
    pod_status = wait_for_pod_status(
        name, namespace, RUNNING_STATE, 90, 0.2
    )  # TODO config
    if pod_status != RUNNING_STATE:
        msg = (
            f"Not running exec commands. Pod {name} {namespace} is not in running state"
        )
        logging.error(msg)
        return msg

    try:
        resp = stream(
            core_v1.connect_get_namespaced_pod_exec,
            name,
            namespace,
            container=container,
            command=exec_command,
            stderr=True,
            stdin=False,
            stdout=True,
            tty=False,
        )

    except ApiException as e:
        if e.status != 404:
            logging.exception(f"exec command {exec_command} resulted with error")
            resp = "error executing commands"

    logging.debug(f"exec command response {resp}")
    return resp


def to_kubernetes_name(name, prefix=""):
    """
    returns a valid and unique kubernetes name based on prefix and name, replacing characters in name as necessary
    see https://kubernetes.io/docs/concepts/overview/working-with-objects/names/
    """
    unique_id = str(time.time()).replace(".", "-")
    safe_name = re.sub("[^0-9a-zA-Z\\-]+", "-", name)
    return f"{prefix}{safe_name}-{unique_id}"[:63]


def parse_kubernetes_datetime(k8s_datetime: str) -> datetime.datetime:
    return datetime.datetime.strptime(k8s_datetime, "%Y-%m-%dT%H:%M:%S%z")


def parse_kubernetes_datetime_to_ms(k8s_datetime: str) -> float:
    return parse_kubernetes_datetime(k8s_datetime).timestamp() * 1000



import os
import uuid
import pytz

INSTALLATION_NAMESPACE = os.environ.get("INSTALLATION_NAMESPACE", "robusta")
TARGET_ID = os.environ.get("RELAY_TARGET_ID", str(uuid.uuid4()))
DISCOVERY_PERIOD_SEC = int(os.environ.get("DISCOVERY_PERIOD_SEC", 90))
SUPABASE_LOGIN_RATE_LIMIT_SEC = int(
    os.environ.get("SUPABASE_LOGIN_RATE_LIMIT_SEC", 900)
)
GRAFANA_RENDERER_URL = os.environ.get(
    "GRAFANA_RENDERER_URL", "http://127.0.0.1:8281/render"
)
SERVICE_UPDATES_CACHE_TTL_SEC = os.environ.get("SERVICE_UPDATES_CACHE_TTL_SEC", 120)
INTERNAL_PLAYBOOKS_ROOT = os.environ.get(
    "INTERNAL_PLAYBOOKS_ROOT", "/app/robusta/core/playbooks/internal"
)
DEFAULT_TIMEZONE = pytz.timezone(os.environ.get("DEFAULT_TIMEZONE", "UTC"))
NUM_EVENT_THREADS = int(os.environ.get("NUM_EVENT_THREADS", 20))



import inspect
import hashlib

# this returns a hash that will change if the function's definition changes
# this is desirable because it lets you warn users if a playbook is called with parameters that
# were generated by a previous version of the playbook. in that case you can ask the user if you should
# call the playbook anyway even though the code that the callback refers to has changed
def get_function_hash(func):
    plaintext = str(inspect.getfullargspec(func)).encode() + func.__code__.co_code
    return hashlib.sha256(plaintext).hexdigest()



# This file contains internal wiring for report callbacks
# Playbook writers don't need to be familiar with this - they should use the API in callbacks.py and not worry about details
import inspect
import logging
from dataclasses import dataclass
from typing import Callable, Any

from pydantic import BaseModel

from ..model.events import BaseEvent
from ...utils.function_hashes import get_function_hash
from ...utils.decorators import doublewrap


class PlaybookCallbackRequest(BaseModel):
    func_name: str
    func_file: str
    func_hash: str
    context: str

    @classmethod
    def create_for_func(cls, func: Callable, context: Any, text: str):
        if func is None:
            raise Exception(
                f"The callback for choice {text} is None. Did you accidentally pass `foo()` as a callback and not `foo`?"
            )
        if not callback_registry.is_callback_in_registry(func):
            raise Exception(
                f"{func} is not a function that was decorated with @on_report_callback or it somehow"
                f" has the wrong version (e.g. multiple functions with the same name were decorated "
                f"with @on_report_callback)"
            )

        return cls(
            func_name=func.__name__,
            func_file=inspect.getsourcefile(func),
            func_hash=get_function_hash(func),
            context=context,
        )


class CallbackRegistry:
    def __init__(self):
        self.callbacks = {}

    def register_callback(self, func: Callable):
        key = self._get_callback_key_for_func(func)
        logging.info(f"inserting callback with key={key}")
        if key in self.callbacks:
            logging.warning(
                f"overriding existing callback in registry; new func={func}"
            )
        self.callbacks[key] = func

    def is_callback_in_registry(self, func: Callable):
        key = self._get_callback_key_for_func(func)
        return key in self.callbacks

    def lookup_callback(self, callback_request: PlaybookCallbackRequest):
        logging.debug(
            f"looking up callback with func_name={callback_request.func_name} and func_file={callback_request.func_file}"
        )
        key = (callback_request.func_name, callback_request.func_file)
        if key not in self.callbacks:
            return None

        func = self.callbacks[key]
        if callback_request.func_hash != get_function_hash(func):
            logging.warning(
                "callback hash doesn't match! calling a different version of the function than the original one!"
            )
        return func

    @staticmethod
    def _get_callback_key_for_func(func: Callable):
        return func.__name__, inspect.getsourcefile(func)


@dataclass
class SinkCallbackEvent(BaseEvent):
    source_channel_name: str = ""
    source_channel_id: str = ""
    source_user_id: str = ""
    source_context: str = ""
    source_message: str = ""


callback_registry = CallbackRegistry()


@doublewrap
def on_sink_callback(func):
    callback_registry.register_callback(func)
    return func



from robusta.api import *
from robusta.core.model.services import ServiceInfo
from robusta.core.discovery.top_service_resolver import TopServiceResolver


@on_kubernetes_any_resource_all_changes
def cluster_discovery_updates(event: KubernetesAnyEvent):
    if (
        event.operation in [K8sOperationType.CREATE, K8sOperationType.UPDATE]
        and event.obj.kind in ["Deployment", "ReplicaSet", "DaemonSet", "StatefulSet"]
        and not event.obj.metadata.ownerReferences
    ):
        TopServiceResolver.add_cached_service(
            ServiceInfo(
                name=event.obj.metadata.name,
                service_type=event.obj.kind,
                namespace=event.obj.metadata.namespace,
            )
        )



import re


class Transformer:
    @staticmethod
    def to_github_markdown(markdown_data: str) -> str:
        """Transform all occurrences of slack markdown, <URL|LINK TEXT>, to github markdown [LINK TEXT](URL)."""
        regex = "<.*?\\|.*?>"
        matches = re.findall(regex, markdown_data)
        if matches:
            matches = [
                match for match in matches if len(match) > 1
            ]  # filter out illegal matches
            for match in matches:
                # take only the data between the first '<' and last '>'
                splits = match[1:-1].split("|")
                if len(splits) == 2:  # don't replace unexpected strings
                    replacement = f"[{splits[1]}]({splits[0]})"
                    markdown_data = markdown_data.replace(match, replacement)
        return markdown_data



import copy
import logging

from .model.playbook_deploy_config import PlaybookDeployConfig
from ..core.sinks.sink_manager import SinkManager
from .model.trigger_params import TriggerParams
from .model.playbook_hash import playbook_hash

from ..core.active_playbooks import get_active_playbooks, get_playbook_inventory
from ..core.model.runner_config import RunnerConfig
from ..core.schedule.scheduler import Scheduler


def clear_active_playbooks():
    get_active_playbooks().clear()


class DeployCommand:
    def __init__(self, deploy_func, func, trigger_params, sinks, action_params=None):
        self.deploy_func = deploy_func
        self.func = func
        self.trigger_params = trigger_params
        self.sinks = sinks
        self.action_params = action_params
        if (
            action_params is not None
            and getattr(action_params, "pre_deploy_func", None) is not None
        ):
            action_params.pre_deploy_func(trigger_params)
        self.playbook_id = playbook_hash(func, trigger_params, action_params)


def deploy_playbook_config(runner_config: RunnerConfig):
    cluster_name = runner_config.global_config.get("cluster_name", "")
    SinkManager.update_sinks_config(runner_config.sinks_config, cluster_name)

    default_sinks = runner_config.global_config.get("sinks", [])
    if not default_sinks:
        logging.warning(
            f"No default sinks defined. By default, actions results are ignored."
        )

    deploy_commands = []
    # TODO we will replace it with a more generic mechanism, as part of the triggers separation task
    # First, we load the internal playbooks, then add the user activated playbooks
    # Order matters. Internal playbooks, should be added first, and run first
    active_playbooks = [PlaybookDeployConfig(name="cluster_discovery_updates")]
    active_playbooks.extend(runner_config.active_playbooks)
    for playbook_config in active_playbooks:
        playbook_definition = get_playbook_inventory().get(playbook_config.name)
        if playbook_definition is None:
            logging.error(
                f"playbook definition not found. skipping. {playbook_config.name}"
            )  # TODO - should we continue, or not run at all??
            continue

        runtime_trigger_params = copy.deepcopy(
            playbook_definition["default_trigger_params"]
        )
        # first override defaults with global config
        runtime_trigger_params = get_merged_global_config(
            runner_config.global_config, runtime_trigger_params
        )
        # then override the result with trigger_params specific config
        runtime_trigger_params = get_merged_config(
            playbook_config.trigger_params, runtime_trigger_params
        )
        playbooks_sinks = (
            playbook_config.sinks
            if playbook_config.sinks is not None
            else default_sinks
        )
        deploy_func = playbook_definition["deploy_func"]
        if playbook_definition["action_params"] is None:
            deploy_commands.append(
                DeployCommand(
                    deploy_func,
                    playbook_definition["func"],
                    runtime_trigger_params,
                    playbooks_sinks,
                )
            )
        else:
            # in case we have params, we have to apply the global config on it as well
            playbook_config.action_params = merge_global_params(
                runner_config.global_config, playbook_config.action_params
            )
            deploy_commands.append(
                DeployCommand(
                    deploy_func,
                    playbook_definition["func"],
                    runtime_trigger_params,
                    playbooks_sinks,
                    playbook_definition["action_params"](
                        **playbook_config.action_params
                    ),
                )
            )

    new_playbook_ids = set(
        [deploy_command.playbook_id for deploy_command in deploy_commands]
    )
    clear_active_playbooks()

    # unschedule playbooks that doesn't exist any more
    Scheduler.unschedule_deleted_playbooks(new_playbook_ids)

    for deploy_command in deploy_commands:
        if deploy_command.action_params is None:
            deploy_command.deploy_func(
                deploy_command.func, deploy_command.trigger_params, deploy_command.sinks
            )
        else:
            deploy_command.deploy_func(
                deploy_command.func,
                deploy_command.trigger_params,
                deploy_command.sinks,
                deploy_command.action_params,
            )


def merge_global_params(global_config: dict, config_params: dict) -> dict:
    merged = global_config.copy()
    merged.update(config_params)
    return merged


def get_merged_global_config(
    global_config: dict, config_defaults: TriggerParams
) -> TriggerParams:
    config_defaults_fields = config_defaults.__fields_set__
    for attribute in global_config.keys():
        if global_config.get(attribute) is not None and hasattr(
            config_defaults_fields, attribute
        ):
            setattr(config_defaults, attribute, getattr(global_config, attribute))
    return config_defaults


def get_merged_config(
    config_overrides: TriggerParams, config_defaults: TriggerParams
) -> TriggerParams:
    for attribute in config_overrides.__fields_set__:
        if getattr(config_overrides, attribute) is not None:
            setattr(config_defaults, attribute, getattr(config_overrides, attribute))
    return config_defaults



from datetime import datetime

from ...core.model.env_vars import DEFAULT_TIMEZONE


class RendererType:
    DATETIME = "DATETIME"


def render_value(renderer: RendererType, value):
    if renderer == RendererType.DATETIME:
        date_value = datetime.fromtimestamp(value / 1000.0)
        return date_value.astimezone(DEFAULT_TIMEZONE).strftime(
            "%b %d, %Y, %I:%M:%S %p"
        )
    raise Exception(f"Unsupported renderer type {renderer}")



import json
import logging

from pydantic import BaseModel
from kafka import KafkaProducer

from ..sink_config import SinkConfigBase
from ...reporting.blocks import Finding, Enrichment, KubernetesDiffBlock, JsonBlock
from ..sink_base import SinkBase


class KafkaSinkConfig(BaseModel):
    kafka_url: str
    topic: str


class KafkaSink(SinkBase):
    def __init__(self, sink_config: SinkConfigBase):
        super().__init__(sink_config)
        config = KafkaSinkConfig(**sink_config.params)
        self.producer = KafkaProducer(bootstrap_servers=config.kafka_url)
        self.topic = config.topic

    def write_finding(self, finding: Finding):
        for enrichment in finding.enrichments:
            self.send_enrichment(
                enrichment,
                finding.subject.name,
                finding.subject.subject_type.value,
                finding.subject.namespace,
            )

    def send_enrichment(
        self,
        enrichment: Enrichment,
        resource_name: str,
        resource_type: str,
        resource_namespace: str,
    ):
        kafka_blocks = [
            block
            for block in enrichment.blocks
            if (isinstance(block, KubernetesDiffBlock) or isinstance(block, JsonBlock))
        ]
        if not kafka_blocks:  # currently supporting sending kafka sink only diffs
            if len(enrichment.blocks) > 0:
                logging.warning(
                    "Kafka sink unsupported enrichment types. Currently only diff/json enrichment is supported"
                )
            return

        message_payload = ""
        for block in kafka_blocks:
            if isinstance(block, KubernetesDiffBlock):
                data = {
                    "resource_name": resource_name,
                    "resource_namespace": resource_namespace,
                    "resource_type": resource_type,
                    "message": f"{resource_type} properties change",
                    "changed_properties": [
                        {
                            "property": ".".join(attribute_diff.path),
                            "old": attribute_diff.other_value,
                            "new": attribute_diff.value,
                        }
                        for attribute_diff in block.diffs
                    ],
                }
                message_payload = json.dumps(data).encode("utf-8")
            elif isinstance(block, JsonBlock):
                message_payload = block.json_str.encode("utf-8")

            self.producer.send(self.topic, value=message_payload)



import os
import random
import subprocess
import time
from distutils.version import StrictVersion
from typing import Optional
import urllib.request
import tarfile

import click_spinner
from zipfile import ZipFile
import yaml
from kubernetes import config


import typer

from .utils import (
    log_title,
    download_file,
    replace_in_file,
    fetch_runner_logs,
    exec_in_robusta_runner,
    get_examples_url,
    PLAYBOOKS_DIR,
)
from .playbooks_cmd import app as playbooks_commands
from .integrations_cmd import app as integrations_commands, get_slack_key


from robusta._version import __version__


app = typer.Typer()
app.add_typer(playbooks_commands, name="playbooks", help="Playbooks commands menu")
app.add_typer(
    integrations_commands, name="integrations", help="Integrations commands menu"
)


def get_runner_url(runner_version=None):
    if runner_version is None:
        runner_version = __version__
    return f"https://gist.githubusercontent.com/robusta-lab/6b809d508dfc3d8d92afc92c7bbbe88e/raw/robusta-{runner_version}.yaml"


CRASHPOD_YAML = "https://gist.githubusercontent.com/robusta-lab/283609047306dc1f05cf59806ade30b6/raw/crashpod.yaml"
SINK_NAME = "sink_name"
SINK_TYPE = "sink_type"
PARAMS = "params"
SINKS = "sinks"
SINKS_CONFIG = "sinks_config"
GLOBAL_CONFIG = "global_config"
CLUSTER_NAME = "cluster_name"
SLACK = "slack"
ROBUSTA = "robusta"


def slack_integration(
    slack_api_key: str, slack_param_file_name: str, slack_channel: str = None
):
    if slack_api_key is None and typer.confirm(
        "do you want to configure slack integration? this is HIGHLY recommended.",
        default=True,
    ):
        slack_api_key = get_slack_key()

        slack_channel = typer.prompt(
            "which slack channel should I send notifications to?"
        )

    if slack_api_key is not None:
        replace_in_file(slack_param_file_name, "<SLACK_API_KEY>", slack_api_key.strip())

    if slack_channel is not None:
        replace_in_file(slack_param_file_name, "<DEFAULT_SLACK_CHANNEL>", slack_channel)


def add_sink(
    global_config: dict, sinks_config: [], sink_name: str, sink_type: str, params: dict
):
    sinks_config.append({SINK_NAME: sink_name, SINK_TYPE: sink_type, PARAMS: params})
    if sink_name not in global_config[SINKS]:
        global_config[SINKS].append(sink_name)


def get_latest_chart_url(chart_entries: list) -> Optional[str]:
    latest_version = StrictVersion("0.0.0")
    latest_version_url = None
    for chart_def in chart_entries:
        chart_version = StrictVersion(chart_def["version"])
        if chart_version > latest_version:
            latest_version = chart_version
            latest_version_url = chart_def["urls"][0]
    return latest_version_url


def get_versioned_chart_url(chart_entries: list, version: str) -> Optional[str]:
    chart_urls = [
        chart["urls"][0] for chart in chart_entries if chart["version"] == version
    ]
    if chart_urls:
        return chart_urls[0]
    return None


def get_chart_url(chart_ver: str = None) -> Optional[str]:
    repo_url = "https://robusta-charts.storage.googleapis.com"
    index_file = urllib.request.urlopen(repo_url + "/index.yaml").read()
    yaml_content = yaml.safe_load(index_file)
    chart_versions = yaml_content.get("entries", {}).get("robusta")
    if chart_ver is not None:
        return get_versioned_chart_url(chart_versions, chart_ver)
    else:
        return get_latest_chart_url(chart_versions)


def get_active_playbooks_yaml(chart_version: str = None):
    chart_url = get_chart_url(chart_version)
    if chart_url is None:
        return None
    local_file_name = "./" + os.path.basename(chart_url)
    try:
        urllib.request.urlretrieve(chart_url, local_file_name)
        with tarfile.open(name=local_file_name, mode="r:gz") as chart_tar_file:
            f = chart_tar_file.extractfile(
                chart_tar_file.getmember("robusta/active_playbooks.yaml")
            )
            if f is not None:
                return yaml.safe_load(f.read())
    finally:
        try:
            os.remove(local_file_name)
        except Exception:
            typer.echo("Failed to delete downloaded chart file", color="red")
    return None


@app.command()
def gen_config(
    base_config_file: str = typer.Option(
        None,
        help="Optional base configuration file. Can be found in Robusta's helm chart. If omitted, the base configuration will be downloaded",
    ),
    use_chart_version: str = typer.Option(
        None,
        help="Optional. Download base configuration from a specific chart version. If omitted, latest chart version will be used",
    ),
):
    """Create runtime configuration file"""
    if base_config_file is None:
        yaml_content = get_active_playbooks_yaml(use_chart_version)
        if yaml_content is None:
            typer.echo(f"Could not find base configuration file", color="red")
            return
    else:
        if not os.path.exists(base_config_file):
            typer.secho(
                f"Base configuration file cannot be found {base_config_file}", fg="red"
            )
            return
        with open(base_config_file, "r") as base:
            yaml_content = yaml.safe_load(base)

    global_config = yaml_content[GLOBAL_CONFIG]

    cluster_name = global_config.get(CLUSTER_NAME)
    if cluster_name is None:
        (all_contexts, current_context) = config.list_kube_config_contexts()
        default_name = (
            current_context.get("name")
            if (current_context and current_context.get("name"))
            else f"cluster_{random.randint(0, 1000000)}"
        )
        cluster_name = typer.prompt(
            "Please specify a unique name for your cluster or press ENTER to use the default",
            default=default_name,
        )
        global_config[CLUSTER_NAME] = cluster_name

    sinks_config = yaml_content[SINKS_CONFIG]

    # Handle slack sink configuration
    slack_sinks = [sink for sink in sinks_config if sink[SINK_TYPE] == SLACK]
    if slack_sinks:
        typer.secho(f"Found slack integration, skipping", fg="green")
    else:
        if typer.confirm(
            "do you want to configure slack integration? this is HIGHLY recommended.",
            default=True,
        ):
            slack_api_key = get_slack_key()

            slack_channel = typer.prompt(
                "which slack channel should I send notifications to?"
            )
            if slack_api_key:
                params = {"api_key": slack_api_key, "slack_channel": slack_channel}
                add_sink(global_config, sinks_config, "slack sink", SLACK, params)

    # Handle robusta sink configuration
    robusta_sinks = [sink for sink in sinks_config if sink[SINK_TYPE] == ROBUSTA]
    if robusta_sinks:
        typer.secho(f"Found robusta integration, skipping", fg="green")
    else:
        if typer.confirm("Would you like to use Robusta UI?"):
            robusta_ui_token = typer.prompt(
                "Please insert your Robusta account token",
                default=None,
            )
            if robusta_ui_token:
                params = {
                    "token": robusta_ui_token,
                }
                add_sink(
                    global_config, sinks_config, "robusta ui sink", ROBUSTA, params
                )

    generated_file = "./active_playbooks_generated.yaml"
    with open(generated_file, "w") as generated:
        yaml.safe_dump(yaml_content, generated, sort_keys=False)
        print(f"Saved configuration to {generated_file}")


def examples_download(
    slack_api_key: str = None,
    slack_channel: str = None,
    cluster_name: str = None,
    use_robusta_ui: bool = False,
    skip_robusta_sink: bool = False,
    robusta_ui_token: str = None,
    url: str = None,
    skip_integrations: bool = False,
):
    """download example playbooks"""
    filename = "example-playbooks.zip"
    if url:
        download_file(url, filename)
    else:
        download_file(get_examples_url(), filename)

    with ZipFile(filename, "r") as zip_file:
        zip_file.extractall()

    if not skip_integrations:
        slack_integration(
            slack_api_key, "playbooks/active_playbooks.yaml", slack_channel
        )

    if cluster_name is None:
        (all_contexts, current_context) = config.list_kube_config_contexts()
        default_name = (
            current_context.get("name")
            if (current_context and current_context.get("name"))
            else f"cluster_{random.randint(0, 1000000)}"
        )
        cluster_name = typer.prompt(
            "Please specify a unique name for your cluster or press ENTER to use the default",
            default=default_name,
        )
    if cluster_name is not None:
        replace_in_file(
            "playbooks/active_playbooks.yaml", "<CLUSTER_NAME>", cluster_name.strip()
        )

    if not skip_robusta_sink and (
        use_robusta_ui or typer.confirm("Would you like to use Robusta UI?")
    ):
        if robusta_ui_token is None:
            robusta_ui_token = typer.prompt(
                "Please insert your Robusta account token",
                default=None,
            )
        if robusta_ui_token is not None:
            replace_in_file(
                "playbooks/active_playbooks.yaml",
                "<ROBUSTA_ACCOUNT_TOKEN>",
                robusta_ui_token.strip(),
            )

        if not skip_robusta_sink:
            replace_in_file(
                "playbooks/active_playbooks.yaml",
                "#<ENABLE_ROBUSTA_SINK>",
                '  - "robusta platform"',
            )

    typer.echo(f"examples downloaded into the {PLAYBOOKS_DIR} directory")


@app.command()
def examples(
    slack_api_key: str = typer.Option(
        None,
        help="Slack api key for Robusta",
    ),
    slack_channel: str = typer.Option(
        None,
        help="Default Slack channel for Robusta",
    ),
    cluster_name: str = typer.Option(
        None,
        help="Unique name for this cluster",
    ),
    use_robusta_ui: bool = typer.Option(
        False,
        help="Use Robusta's ui?",
    ),
    skip_robusta_sink: bool = typer.Option(
        False,
        help="Enable Robusta sink?",
    ),
    robusta_ui_token: str = typer.Option(
        None,
        help="Robusta UI account token",
    ),
    url: str = typer.Option(
        None,
        help="Deploy Robusta playbooks from a given url instead of using the latest version",
    ),
    skip_integrations: bool = typer.Option(
        False,
        help="Skip integrations configuration",
    ),
):
    """Download playbooks code and configuration defaults"""
    examples_download(
        slack_api_key,
        slack_channel,
        cluster_name,
        use_robusta_ui,
        skip_robusta_sink,
        robusta_ui_token,
        url,
        skip_integrations,
    )


@app.command()
def playground():
    """open a python playground - useful when writing playbooks"""
    typer.echo(
        "this command is temporarily disabled due to recent changes to python:3.8-slim"
    )
    # exec_in_robusta_runner("socat readline unix-connect:/tmp/manhole-1")


@app.command()
def version():
    """show the version of the local robusta-cli"""
    if __version__ == "0.0.0":
        typer.echo("running with development version from git")
    else:
        typer.echo(f"version {__version__}")


@app.command()
def demo(
    namespace: str = typer.Option(
        "robusta",
        help="Demo Robusta on the specified custom namespace",
    ),
):
    """deliberately deploy a crashing pod to kubernetes so you can test robusta's response"""
    log_title("Deploying a crashing pod to kubernetes...")
    subprocess.check_call(f"kubectl apply -f {CRASHPOD_YAML}", shell=True)
    log_title(
        "In ~30 seconds you should receive a slack notification on a crashing pod"
    )
    time.sleep(60)
    subprocess.check_call(
        f"kubectl delete -n {namespace} deployment crashpod", shell=True
    )
    log_title("Done!")


if __name__ == "__main__":
    app()



import base64
import json
import logging
import threading
import time
import traceback
import uuid

from typing import List, Dict, Any

from supabase_py.lib.auth_client import SupabaseAuthClient

from ...transformer import Transformer
from ....discovery.top_service_resolver import TopServiceResolver
from ....model.services import ServiceInfo
from ....reporting.blocks import (
    Finding,
    Enrichment,
    MarkdownBlock,
    KubernetesDiffBlock,
    DividerBlock,
    FileBlock,
    HeaderBlock,
    CallbackBlock,
    ListBlock,
    TableBlock,
)
from ....model.env_vars import TARGET_ID, SUPABASE_LOGIN_RATE_LIMIT_SEC
from ....reporting.callbacks import PlaybookCallbackRequest
from supabase_py import Client

SERVICES_TABLE = "Services"
EVIDENCE_TABLE = "Evidence"
ISSUES_TABLE = "Issues"


class RobustaAuthClient(SupabaseAuthClient):
    def _set_timeout(*args, **kwargs):
        """Set timer task"""
        # _set_timeout isn't implemented in gotrue client. it's required for the jwt refresh token timer task
        # https://github.com/supabase/gotrue-py/blob/49c092e3a4a6d7bb5e1c08067a4c42cc2f74b5cc/gotrue/client.py#L242
        # callback, timeout_ms
        threading.Timer(args[2] / 1000, args[1]).start()


class RobustaClient(Client):
    def _get_auth_headers(self) -> Dict[str, str]:
        auth = getattr(self, "auth", None)
        session = auth.current_session if auth else None
        if session and session["access_token"]:
            access_token = auth.session()["access_token"]
        else:
            access_token = self.supabase_key

        headers: Dict[str, str] = {
            "apiKey": self.supabase_key,
            "Authorization": f"Bearer {access_token}",
        }
        return headers

    @staticmethod
    def _init_supabase_auth_client(
        auth_url: str,
        supabase_key: str,
        detect_session_in_url: bool,
        auto_refresh_token: bool,
        persist_session: bool,
        local_storage: Dict[str, Any],
        headers: Dict[str, str],
    ) -> RobustaAuthClient:
        """Creates a wrapped instance of the GoTrue Client."""
        return RobustaAuthClient(
            url=auth_url,
            auto_refresh_token=auto_refresh_token,
            detect_session_in_url=detect_session_in_url,
            persist_session=persist_session,
            local_storage=local_storage,
            headers=headers,
        )


class SupabaseDal:
    def __init__(
        self,
        url: str,
        key: str,
        account_id: str,
        email: str,
        password: str,
        sink_name: str,
        cluster_name: str,
    ):
        self.url = url
        self.key = key
        self.account_id = account_id
        self.cluster = cluster_name
        self.client = RobustaClient(url, key)
        self.email = email
        self.password = password
        self.sign_in_time = 0
        self.sign_in()
        self.target_id = TARGET_ID
        self.sink_name = sink_name

    def to_issue(self, finding: Finding):
        return {
            "id": str(finding.id),
            "title": finding.title,
            "description": finding.description,
            "source": finding.source.value,
            "aggregation_key": finding.aggregation_key,
            "failure": finding.failure,
            "finding_type": finding.finding_type.value,
            "category": finding.category,
            "priority": finding.severity.name,
            "subject_type": finding.subject.subject_type.value,
            "subject_name": finding.subject.name,
            "subject_namespace": finding.subject.namespace,
            "service_key": TopServiceResolver.guess_service_key(
                finding.subject.name, finding.subject.namespace
            ),
            "cluster": self.cluster,
            "account_id": self.account_id,
        }

    def to_evidence(self, finding_id: uuid, enrichment: Enrichment) -> Dict[Any, Any]:
        structured_data = []
        for block in enrichment.blocks:
            if isinstance(block, MarkdownBlock):
                if not block.text:
                    continue
                structured_data.append(
                    {
                        "type": "markdown",
                        "data": Transformer.to_github_markdown(block.text),
                    }
                )
            elif isinstance(block, DividerBlock):
                structured_data.append({"type": "divider"})
            elif isinstance(block, FileBlock):
                last_dot_idx = block.filename.rindex(".")
                structured_data.append(
                    {
                        "type": block.filename[last_dot_idx + 1 :],
                        "data": str(base64.b64encode(block.contents)),
                    }
                )
            elif isinstance(block, HeaderBlock):
                structured_data.append({"type": "header", "data": block.text})
            elif isinstance(block, ListBlock):
                structured_data.append({"type": "list", "data": block.items})
            elif isinstance(block, TableBlock):
                structured_data.append(
                    {
                        "type": "table",
                        "data": {
                            "headers": block.headers,
                            "rows": [row for row in block.rows],
                            "column_renderers": block.column_renderers,
                        },
                    }
                )
            elif isinstance(block, KubernetesDiffBlock):
                structured_data.append(
                    {
                        "type": "diff",
                        "data": {
                            "old": block.old,
                            "new": block.new,
                        },
                    }
                )
            elif isinstance(block, CallbackBlock):
                context = block.context.copy()
                context["target_id"] = self.target_id
                context["sink_name"] = self.sink_name
                callbacks = []
                for (text, callback) in block.choices.items():
                    callbacks.append(
                        {
                            "text": text,
                            "callback": PlaybookCallbackRequest.create_for_func(
                                callback, json.dumps(context), text
                            ).json(),
                        }
                    )

                structured_data.append({"type": "callbacks", "data": callbacks})
            else:
                logging.error(
                    f"cannot convert block of type {type(block)} to robusta platform format block: {block}"
                )
                continue  # no reason to crash the entire report

        return {
            "issue_id": str(finding_id),
            "file_type": "structured_data",
            "data": json.dumps(structured_data),
            "account_id": self.account_id,
        }

    def persist_finding(self, finding: Finding):
        for enrichment in finding.enrichments:
            res = (
                self.client.table(EVIDENCE_TABLE)
                .insert(self.to_evidence(finding.id, enrichment))
                .execute()
            )
            if res.get("status_code") != 201:
                logging.error(
                    f"Failed to persist finding {finding.id} enrichment {enrichment} error: {res.get('data')}"
                )

        res = self.client.table(ISSUES_TABLE).insert(self.to_issue(finding)).execute()
        if res.get("status_code") != 201:
            logging.error(
                f"Failed to persist finding {finding.id} error: {res.get('data')}"
            )
            self.handle_supabase_error()

    def to_service(self, service: ServiceInfo) -> Dict[Any, Any]:
        return {
            "name": service.name,
            "type": service.service_type,
            "namespace": service.namespace,
            "classification": service.classification,
            "cluster": self.cluster,
            "account_id": self.account_id,
            "deleted": service.deleted,
            "service_key": service.get_service_key(),
            "update_time": "now()",
        }

    def persist_service(self, service: ServiceInfo):
        db_service = self.to_service(service)
        res = (
            self.client.table(SERVICES_TABLE).insert(db_service, upsert=True).execute()
        )
        if res.get("status_code") not in [200, 201]:
            logging.error(
                f"Failed to persist service {service} error: {res.get('data')}"
            )
            self.handle_supabase_error()

    def get_active_services(self) -> List[ServiceInfo]:
        res = (
            self.client.table(SERVICES_TABLE)
            .select("name", "type", "namespace", "classification")
            .filter("account_id", "eq", self.account_id)
            .filter("cluster", "eq", self.cluster)
            .filter("deleted", "eq", False)
            .execute()
        )
        if res.get("status_code") not in [200]:
            msg = f"Failed to get existing services (supabase) error: {res.get('data')}"
            logging.error(msg)
            self.handle_supabase_error()
            raise Exception(msg)

        return [
            ServiceInfo(
                name=service["name"],
                service_type=service["type"],
                namespace=service["namespace"],
                classification=service["classification"],
            )
            for service in res.get("data")
        ]

    def sign_in(self):
        if time.time() > self.sign_in_time + SUPABASE_LOGIN_RATE_LIMIT_SEC:
            logging.info("Supabase dal login")
            self.sign_in_time = time.time()
            self.client.auth.sign_in(email=self.email, password=self.password)

    def handle_supabase_error(self):
        """Workaround for Gotrue bug in refresh token."""
        # If there's an error during refresh token, no new refresh timer task is created, and the client remains not authenticated for good
        # When there's an error connecting to supabase server, we will re-login, to re-authenticate the session.
        # Adding rate-limiting mechanism, not to login too much because of other errors
        # https://github.com/supabase/gotrue-py/issues/9
        try:
            self.sign_in()
        except Exception as e:
            logging.error("Failed to signin on error", traceback.print_exc())



import inspect
import logging
from typing import Callable
from collections import defaultdict
from sys import exc_info

import better_exceptions  # type: ignore
from ..core.model.cloud_event import CloudEvent


class PlaybookWrapper:
    def __init__(self, wrapper, playbook_id):
        self.wrapper = wrapper
        self.playbook_id = playbook_id


active_playbooks = defaultdict(list)  # maps trigger types to active playbooks
playbook_inventory = {}


def get_playbook_inventory():
    return playbook_inventory


def get_active_playbooks():
    return active_playbooks


def get_function_params_class(func: Callable):
    """Inspects a playbook function's signature and returns the type of the param class if it exists"""
    func_signature = inspect.signature(func)
    if len(func_signature.parameters) == 1:
        return None
    parameter_name = list(func_signature.parameters)[1]
    return func_signature.parameters[parameter_name].annotation


def register_playbook(func, deploy_func, default_trigger_params):
    get_playbook_inventory()[func.__name__] = {
        "func": func,
        "default_trigger_params": default_trigger_params,
        "deploy_func": deploy_func,
        "action_params": get_function_params_class(func),
    }
    func.__playbook = playbook_inventory[func.__name__]


def clear_playbook_inventory():
    playbook_inventory.clear()


def activate_playbook(trigger_type, wrapper, func, playbook_id):
    logging.info(f"adding handler {func} playbook_id {playbook_id}")
    active_playbooks[trigger_type.name].append(PlaybookWrapper(wrapper, playbook_id))


def run_playbooks(cloud_event: CloudEvent):
    # TODO: Ideally we would do the conversion to a concrete event class here so that we pass the same event
    # object to all playbooks that are triggered and they can each add stuff to the reporting blocks
    description = cloud_event.data["description"].replace("\n", "")
    logging.debug(f"received cloud event: {description}")
    handlers = active_playbooks[cloud_event.type]
    logging.debug(f"relevant handlers: {handlers}")
    for playbook_wrapper in handlers:
        try:
            playbook_wrapper.wrapper(cloud_event)
        except Exception as e:
            _, _, traceback = exc_info()
            msg = "\n".join(
                better_exceptions.format_exception(e.__class__, e, traceback)
            )
            logging.exception(f"got exception running handler: {msg}")



import os
import threading
import time, logging
from collections import defaultdict

from typing import List

from ...core.persistency.scheduled_jobs_states_dal import SchedulerDal
from ...core.schedule.model import (
    JobStatus,
    ScheduledJob,
    DynamicDelayRepeat,
    SchedulingInfo,
)

INITIAL_SCHEDULE_DELAY_SEC = os.environ.get("INITIAL_SCHEDULE_DELAY_SEC", 5)


class Scheduler:
    scheduled_jobs = defaultdict(None)
    registered_runnables = {}
    dal = None

    @classmethod
    def register_task(cls, runnable_name: str, func):
        cls.registered_runnables[runnable_name] = func

    @classmethod
    def init_scheduler(cls):
        cls.dal = SchedulerDal()
        # schedule standalone tasks
        for job in cls.__get_standalone_jobs():
            logging.info(f"Scheduling standalone task {job.job_id}")
            cls.schedule_job(job)

    @classmethod
    def schedule_job(cls, job: ScheduledJob):
        if job.replace_existing:
            cls.__remove_scheduler_job(job.job_id)
            saved_job = None  # Don't load the existing state. Will be overridden with a new state
        else:
            if cls.__is_scheduled(job.job_id):
                logging.info(f"job {job.job_id} already scheduled")
                return  # job is already scheduled, no need to re-schedule. (this is a reload playbooks scenario)
            saved_job = cls.dal.get_scheduled_job(job.job_id)

        if saved_job is None:  # save new job
            cls.dal.save_scheduled_job(job)
            saved_job = job
        elif saved_job.state.job_status == JobStatus.DONE:
            logging.info(
                f"Scheduled job already done. Skipping scheduling. job {saved_job.job_id}"
            )
            return

        next_delay = cls.__calc_job_delay_for_next_run(saved_job)
        logging.info(
            f"scheduling job {saved_job.job_id} params {saved_job.scheduling_params} will run in {next_delay}"
        )
        cls.__schedule_job_internal(
            next_delay, saved_job.job_id, cls.__on_task_execution, {"job": saved_job}
        )

    @classmethod
    def unschedule_deleted_playbooks(cls, active_playbook_ids: set):
        for job in cls.dal.list_scheduled_jobs():
            if job.standalone_task:
                continue  # standalone tasks shouldn't be removed on reload
            if job.job_id not in active_playbook_ids:
                logging.info(f"unscheduling deleted playbook {job.job_id}")
                cls.__unschedule_job(job.job_id)

    @classmethod
    def __on_task_execution(cls, job: ScheduledJob):
        logging.info(f"running scheduled job {job.job_id}")

        if job.state.job_status == JobStatus.NEW:
            job.state.job_status = JobStatus.RUNNING
        job.state.last_exec_time_sec = round(time.time())

        func = cls.registered_runnables.get(job.runnable_name)
        if not func:
            logging.error(f"Scheduled runnable name not registered {job.runnable_name}")
            cls.__on_job_done(job)
            return

        try:
            func(
                runnable_params=job.runnable_params,
                schedule_info=SchedulingInfo(execution_count=job.state.exec_count),
            )
        except Exception:
            logging.exception(
                f"failed to execute runnable {job.runnable_name}. job_id {job.job_id} exec_count {job.state.exec_count}"
            )

        job.state.exec_count += 1
        if cls.__is_job_done(job):
            cls.__on_job_done(job)
        else:
            cls.dal.save_scheduled_job(job)
            next_delay = cls.__calc_job_delay_for_next_run(job)
            cls.__schedule_job_internal(
                next_delay,
                job.job_id,
                cls.__on_task_execution,
                {"job": job},
            )

    @classmethod
    def __get_standalone_jobs(cls) -> List[ScheduledJob]:
        return [job for job in cls.dal.list_scheduled_jobs() if job.standalone_task]

    @classmethod
    def __on_job_done(cls, job: ScheduledJob):
        job.state.job_status = JobStatus.DONE
        # need to persist jobs state before unscheduling the job. (to avoid race condition, on configuration reload)
        if job.standalone_task:
            cls.dal.del_scheduled_job(job.job_id)
        else:
            cls.dal.save_scheduled_job(job)
        del cls.scheduled_jobs[job.job_id]
        logging.info(
            f"Scheduled job done. job_id {job.job_id} executions {job.state.exec_count}"
        )

    @classmethod
    def __schedule_job_internal(cls, delay, job_id, func, kwargs):
        job = threading.Timer(delay, func, kwargs=kwargs)
        cls.scheduled_jobs[job_id] = job
        job.start()

    @classmethod
    def __remove_scheduler_job(cls, job_id):
        job = cls.scheduled_jobs.get(job_id)
        if job is not None:
            job.cancel()
            del cls.scheduled_jobs[job_id]

    @classmethod
    def __unschedule_job(cls, job_id):
        cls.__remove_scheduler_job(job_id)
        cls.dal.del_scheduled_job(job_id)

    @classmethod
    def __is_scheduled(cls, job_id):
        return cls.scheduled_jobs.get(job_id) is not None

    @classmethod
    def __is_job_done(cls, job: ScheduledJob) -> bool:
        if isinstance(job.scheduling_params, DynamicDelayRepeat):
            return job.state.exec_count == len(job.scheduling_params.delay_periods)
        else:  # default, FIXED_DELAY_REPEAT
            return job.state.exec_count == job.scheduling_params.repeat

    @classmethod
    def __calc_job_delay_for_next_run(cls, job: ScheduledJob):
        if job.state.job_status == JobStatus.NEW:
            if isinstance(job.scheduling_params, DynamicDelayRepeat):
                return job.scheduling_params.delay_periods[0]
            else:
                return INITIAL_SCHEDULE_DELAY_SEC

        if isinstance(job.scheduling_params, DynamicDelayRepeat):
            next_delay = job.scheduling_params.delay_periods[job.state.exec_count]
        else:  # FIXED_DELAY_REPEAT type
            next_delay = job.scheduling_params.seconds_delay

        return max(
            job.state.last_exec_time_sec + next_delay - round(time.time()),
            INITIAL_SCHEDULE_DELAY_SEC,
        )



# Relay target id
from enum import Enum

# sink types
class SinkType(Enum):
    ROBUSTA = "robusta"
    SLACK = "slack"
    KAFKA = "kafka"
    DATADOG = "datadog"


class FindingType(Enum):
    ISSUE = "issue"
    CONF_CHANGE = "configuration_change"
    HEALTH_CHECK = "health_check"
    REPORT = "report"


# Finding sources
class FindingSource(Enum):
    NONE = None  # empty default
    KUBERNETES_API_SERVER = "kubernetes_api_server"
    PROMETHEUS = "prometheus"
    MANUAL = "manual"
    CALLBACK = "callback"


# Finding subject types
class FindingSubjectType(Enum):
    TYPE_NONE = None  # empty default
    TYPE_DEPLOYMENT = "deployment"
    TYPE_NODE = "node"
    TYPE_POD = "pod"
    TYPE_JOB = "job"
    TYPE_DAEMONSET = "daemonset"


# Annotations
class SlackAnnotations:
    UNFURL = "unfurl"
    ATTACHMENT = "attachment"



# This file was autogenerated. Do not edit.

from hikaru.model.rel_1_16.v2beta2 import *
from ...custom_models import RobustaPod, RobustaDeployment


KIND_TO_MODEL_CLASS = {
    "Pod": RobustaPod,
    "ReplicaSet": ReplicaSet,
    "DaemonSet": DaemonSet,
    "Deployment": RobustaDeployment,
    "StatefulSet": StatefulSet,
    "Service": Service,
    "ConfigMap": ConfigMap,
    "Event": Event,
    "HorizontalPodAutoscaler": HorizontalPodAutoscaler,
    "Node": Node,
}



# This file was autogenerated. Do not edit.

from ....utils.decorators import doublewrap
from ..base_triggers import register_k8s_playbook, register_k8s_any_playbook
from ....core.model.k8s_operation_type import K8sOperationType


# Pod Triggers
@doublewrap
def on_pod_create(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "Pod",
        K8sOperationType.CREATE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_pod_update(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "Pod",
        K8sOperationType.UPDATE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_pod_delete(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "Pod",
        K8sOperationType.DELETE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_pod_all_changes(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func, "Pod", None, name_prefix=name_prefix, namespace_prefix=namespace_prefix
    )


# ReplicaSet Triggers
@doublewrap
def on_replicaset_create(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "ReplicaSet",
        K8sOperationType.CREATE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_replicaset_update(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "ReplicaSet",
        K8sOperationType.UPDATE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_replicaset_delete(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "ReplicaSet",
        K8sOperationType.DELETE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_replicaset_all_changes(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "ReplicaSet",
        None,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


# DaemonSet Triggers
@doublewrap
def on_daemonset_create(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "DaemonSet",
        K8sOperationType.CREATE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_daemonset_update(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "DaemonSet",
        K8sOperationType.UPDATE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_daemonset_delete(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "DaemonSet",
        K8sOperationType.DELETE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_daemonset_all_changes(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "DaemonSet",
        None,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


# Deployment Triggers
@doublewrap
def on_deployment_create(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "Deployment",
        K8sOperationType.CREATE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_deployment_update(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "Deployment",
        K8sOperationType.UPDATE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_deployment_delete(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "Deployment",
        K8sOperationType.DELETE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_deployment_all_changes(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "Deployment",
        None,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


# StatefulSet Triggers
@doublewrap
def on_statefulset_create(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "StatefulSet",
        K8sOperationType.CREATE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_statefulset_update(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "StatefulSet",
        K8sOperationType.UPDATE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_statefulset_delete(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "StatefulSet",
        K8sOperationType.DELETE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_statefulset_all_changes(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "StatefulSet",
        None,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


# Service Triggers
@doublewrap
def on_service_create(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "Service",
        K8sOperationType.CREATE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_service_update(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "Service",
        K8sOperationType.UPDATE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_service_delete(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "Service",
        K8sOperationType.DELETE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_service_all_changes(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "Service",
        None,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


# ConfigMap Triggers
@doublewrap
def on_configmap_create(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "ConfigMap",
        K8sOperationType.CREATE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_configmap_update(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "ConfigMap",
        K8sOperationType.UPDATE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_configmap_delete(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "ConfigMap",
        K8sOperationType.DELETE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_configmap_all_changes(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "ConfigMap",
        None,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


# Event Triggers
@doublewrap
def on_event_create(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "Event",
        K8sOperationType.CREATE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_event_update(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "Event",
        K8sOperationType.UPDATE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_event_delete(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "Event",
        K8sOperationType.DELETE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_event_all_changes(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func, "Event", None, name_prefix=name_prefix, namespace_prefix=namespace_prefix
    )


# HorizontalPodAutoscaler Triggers
@doublewrap
def on_horizontalpodautoscaler_create(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "HorizontalPodAutoscaler",
        K8sOperationType.CREATE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_horizontalpodautoscaler_update(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "HorizontalPodAutoscaler",
        K8sOperationType.UPDATE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_horizontalpodautoscaler_delete(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "HorizontalPodAutoscaler",
        K8sOperationType.DELETE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_horizontalpodautoscaler_all_changes(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "HorizontalPodAutoscaler",
        None,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


# Node Triggers
@doublewrap
def on_node_create(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "Node",
        K8sOperationType.CREATE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_node_update(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "Node",
        K8sOperationType.UPDATE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_node_delete(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func,
        "Node",
        K8sOperationType.DELETE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_node_all_changes(func, name_prefix="", namespace_prefix=""):
    return register_k8s_playbook(
        func, "Node", None, name_prefix=name_prefix, namespace_prefix=namespace_prefix
    )


# Kubernetes Any Triggers
@doublewrap
def on_kubernetes_any_resource_create(func, name_prefix="", namespace_prefix=""):
    return register_k8s_any_playbook(
        func,
        K8sOperationType.CREATE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_kubernetes_any_resource_update(func, name_prefix="", namespace_prefix=""):
    return register_k8s_any_playbook(
        func,
        K8sOperationType.UPDATE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_kubernetes_any_resource_delete(func, name_prefix="", namespace_prefix=""):
    return register_k8s_any_playbook(
        func,
        K8sOperationType.DELETE,
        name_prefix=name_prefix,
        namespace_prefix=namespace_prefix,
    )


@doublewrap
def on_kubernetes_any_resource_all_changes(func, name_prefix="", namespace_prefix=""):
    return register_k8s_any_playbook(
        func, None, name_prefix=name_prefix, namespace_prefix=namespace_prefix
    )



import glob
import importlib.util
import logging
import os
import subprocess
import sys
import threading
from typing import Optional

import yaml

from ..core.model.env_vars import INTERNAL_PLAYBOOKS_ROOT
from ..integrations.git.git_repo import GitRepoManager
from ..core.active_playbooks import clear_playbook_inventory
from ..core.triggers import deploy_playbook_config, RunnerConfig
from ..utils.directory_watcher import DirWatcher


class ConfigLoader:

    # the structure on disk is:
    # root_playbook_path/
    # |- playbook_dir1
    #    |--- playbook1.py
    #    |--- playbook2.py
    #    |--- requirements.txt
    # |- playbook_dir2
    #    |--- ...
    def __init__(self, config_file_path: str, root_playbook_path: str):
        self.config_file_path = config_file_path
        self.root_playbook_path = root_playbook_path
        self.reload_lock = threading.Lock()
        self.watcher = DirWatcher(self.root_playbook_path, self.__reload)
        self.conf_watcher = DirWatcher(
            os.path.dirname(self.config_file_path), self.__reload
        )
        self.__reload("initialization")

    def close(self):
        self.watcher.stop_watcher()
        self.conf_watcher.stop_watcher()

    def __reload(self, changed_location):
        logging.info(f"Reloading configuration due to change on {changed_location}")
        with self.reload_lock:
            try:
                # TODO: there is a race condition here where we can lose events if they arrive while we are reloading
                # even if the playbook that should handle those events was active in both versions
                # We should ultimately fix this by replacing clear_playbook_inventory() with an atomic call to
                # set_playbook_inventory(playbooks) - but to do so we first have to change the way playbook registration
                # works. In the new design, playbooks should be found instead of registering themselves.
                runner_config = self.__load_config(self.config_file_path)
                if runner_config is None:
                    return

                playbook_directories = [
                    os.path.join(self.root_playbook_path, path)
                    for path in runner_config.playbook_sets
                ]

                clear_playbook_inventory()
                self.__load_playbook_directory(INTERNAL_PLAYBOOKS_ROOT)
                for playbook_dir in playbook_directories:
                    self.__load_playbook_directory(playbook_dir)

                deploy_playbook_config(runner_config)
                # clear git repos, so it would be re-initialized
                GitRepoManager.clear_git_repos()

            except Exception as e:
                logging.exception(
                    f"unknown error reloading playbooks. will try again when they next change. exception={e}"
                )

    @classmethod
    def __load_config(cls, config_file_path) -> Optional[RunnerConfig]:
        if not os.path.exists(config_file_path):
            logging.warning(
                f"config file not found at {config_file_path} - not configuring any playbooks."
            )
            return None

        logging.info(f"loading config {config_file_path}")
        with open(config_file_path) as file:
            yaml_content = yaml.safe_load(file)
            return RunnerConfig(**yaml_content)

    @classmethod
    def __load_playbook_directory(cls, playbook_dir):
        logging.info(f"Loading playbooks sources from {playbook_dir}")
        if not os.path.exists(playbook_dir):
            logging.error(f"playbooks directory not found: {playbook_dir}")
            return

        # we add playbook directories to sys.path so that playbooks can do relative imports from one another.
        # e.g. `from other_playbook_file import shared_function`
        if playbook_dir not in sys.path:
            sys.path.append(playbook_dir)

        cls.__install_requirements(os.path.join(playbook_dir, "requirements.txt"))
        python_files = glob.glob(f"{playbook_dir}/*.py")
        if len(python_files) == 0:
            logging.warning(f"no playbook scripts to load in directory {playbook_dir}")
            return

        for script in python_files:
            try:
                logging.info(f"loading playbooks from file {script}")
                filename = os.path.basename(script)
                (module_name, ext) = os.path.splitext(filename)
                spec = importlib.util.spec_from_file_location(module_name, script)
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
            except Exception as e:
                logging.error(
                    f"error loading playbooks from file {script}. exception={e}"
                )

        logging.info(f"{len(python_files)} playbooks loaded from {playbook_dir}")

    @classmethod
    def __install_requirements(cls, requirements_path):
        if os.path.exists(requirements_path):
            logging.info(f"installing requirements file {requirements_path}")
            subprocess.check_call(
                [sys.executable, "-m", "pip", "install", "-r", requirements_path]
            )
            logging.info("requirements installed")
        else:
            logging.info(
                f"not installing requirements as file {requirements_path} doesn't exist"
            )



from enum import Enum

from pydantic import BaseModel
from typing import List, Union


class JobStatus(Enum):
    NEW = 1
    RUNNING = 2
    DONE = 3


class FixedDelayRepeat(BaseModel):
    repeat: int
    seconds_delay: int


class DynamicDelayRepeat(BaseModel):
    delay_periods: List[int]


class JobState(BaseModel):
    exec_count: int = 0
    job_status: JobStatus = JobStatus.NEW
    last_exec_time_sec: int = 0


class ScheduledJob(BaseModel):
    job_id: str
    runnable_name: str
    runnable_params: dict
    state: JobState
    scheduling_params: Union[FixedDelayRepeat, DynamicDelayRepeat]
    replace_existing: bool = False
    standalone_task: bool = False  # standalone task, is a task, that once fired, will run until it ends, as opposed to non standalone tasks,
    # that are tied to the generating playbook lifecycle


class SchedulingInfo(BaseModel):
    execution_count: int



import logging
import threading
import time

from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

WAIT_SEC = 2


class FsChangeHandler(FileSystemEventHandler):
    """reload playbooks on change."""

    def __init__(self, mark_change):
        super().__init__()
        self.mark_change = mark_change

    def on_moved(self, event):
        self.mark_change()

    def on_created(self, event):
        self.mark_change()

    def on_deleted(self, event):
        self.mark_change()

    def on_modified(self, event):
        self.mark_change()


class DirWatcher:
    def __init__(self, path_to_watch, reload_configuration):
        self.active = True
        self.change_detected = False
        self.path_to_watch = path_to_watch
        self.reload_configuration = reload_configuration

        self.watch_thread = threading.Thread(target=self.watch, name="config-watcher")
        self.watch_thread.start()

        logging.info(f"watching dir {path_to_watch} for custom playbooks changes")

    def watch(self):
        observer = Observer()
        fs_change_handler = FsChangeHandler(self.mark_change)
        observer.schedule(fs_change_handler, self.path_to_watch, recursive=True)
        observer.start()
        try:
            while True:
                if not self.active:
                    break
                time.sleep(WAIT_SEC)

                if self.change_detected:
                    time.sleep(
                        WAIT_SEC
                    )  # once we detected a change, we wait a safety period to make sure all the changes under this 'bulk' are finished
                    self.change_detected = False
                    try:
                        self.reload_configuration(self.path_to_watch)
                    except Exception as e:  # in case we have an error while trying to reload, we want the watch thread to stay alive
                        logging.exception("failed to reload configuration")
        finally:
            observer.stop()
            observer.join()

    def stop_watcher(self):
        self.active = False

    def mark_change(self):
        self.change_detected = True



import re
from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional, Dict, Any, Union
from pydantic import BaseModel

from hikaru.model import Node, DaemonSet

from ...core.model.events import BaseEvent
from ..kubernetes.custom_models import RobustaPod, RobustaDeployment, RobustaJob


# for parsing incoming data
class PrometheusAlert(BaseModel):
    endsAt: datetime
    generatorURL: str
    startsAt: datetime
    fingerprint: Optional[str] = ""
    status: str
    labels: Dict[Any, Any]
    annotations: Dict[Any, Any]


# for parsing incoming data
class PrometheusEvent(BaseModel):
    alerts: List[PrometheusAlert] = []
    description: str
    externalURL: str
    groupKey: str
    version: str
    commonAnnotations: Optional[Dict[Any, Any]] = None
    commonLabels: Optional[Dict[Any, Any]] = None
    groupLabels: Optional[Dict[Any, Any]] = None
    receiver: str
    status: str


# everything here needs to be optional due to annoying subtleties regarding dataclass inheritance
# see explanation in the code for BaseEvent
@dataclass
class PrometheusKubernetesAlert(BaseEvent):
    alert: Optional[PrometheusAlert] = None
    alert_name: Optional[str] = None
    alert_severity: Optional[str] = None
    node: Optional[Node] = None
    pod: Optional[RobustaPod] = None
    deployment: Optional[RobustaDeployment] = None
    job: Optional[RobustaJob] = None
    daemonset: Optional[DaemonSet] = None

    def get_title(self) -> str:
        annotations = self.alert.annotations
        if annotations.get("summary"):
            return f'{annotations["summary"]}'
        else:
            return self.alert.labels.get("alertname", "")

    def get_description(self) -> str:
        annotations = self.alert.annotations
        clean_description = ""
        if annotations.get("description"):
            # remove "LABELS = map[...]" from the description as we already add a TableBlock with labels
            clean_description = re.sub(
                r"LABELS = map\[.*\]$", "", annotations["description"]
            )
        return clean_description



import logging
import threading
import time
from collections import defaultdict
from typing import List

from pydantic.main import BaseModel

from ..model.env_vars import SERVICE_UPDATES_CACHE_TTL_SEC
from ..model.services import ServiceInfo


class CachedServiceInfo(BaseModel):
    service_info: ServiceInfo
    event_time: float


class TopServiceResolver:
    __recent_service_updates = {}
    __namespace_to_service = defaultdict(list)
    __cached_updates_lock = threading.Lock()

    @classmethod
    def store_cached_services(cls, services: List[ServiceInfo]):
        new_store = defaultdict(list)
        for service in services:
            new_store[service.namespace].append(service)

        # The services are stored periodically, after reading it from the API server. If, between reads
        # new services are added, they will be missing from the cache. So, in addition to the periodic read, we
        # update the cache from listening to add/update API server events.
        # append recent updates, to avoid race conditions between events and api server read
        with cls.__cached_updates_lock:
            recent_updates_keys = list(cls.__recent_service_updates.keys())
            for service_key in recent_updates_keys:
                recent_update = cls.__recent_service_updates[service_key]
                if (
                    time.time() - recent_update.event_time
                    > SERVICE_UPDATES_CACHE_TTL_SEC
                ):
                    del cls.__recent_service_updates[service_key]
                else:
                    new_store[recent_update.service_info.namespace].append(
                        recent_update.service_info
                    )

        cls.__namespace_to_service = new_store

    # TODO remove this guess function
    # temporary try to guess who the owner service is.
    @classmethod
    def guess_service_key(cls, name: str, namespace: str) -> str:
        for cached_service in cls.__namespace_to_service[namespace]:
            if name.startswith(cached_service.name):
                return cached_service.get_service_key()
        return ""

    @classmethod
    def add_cached_service(cls, service: ServiceInfo):
        cls.__namespace_to_service[service.namespace].append(service)
        with cls.__cached_updates_lock:
            cls.__recent_service_updates[service.get_service_key()] = CachedServiceInfo(
                service_info=service, event_time=time.time()
            )



from hikaru import DiffDetail
from typing import List


# TODO: filter out all the managed fields crap
def is_matching_diff(diff_detail: DiffDetail, fields_to_monitor: List[str]) -> bool:
    return any(
        substring in diff_detail.formatted_path for substring in fields_to_monitor
    )



from typing import List, Optional

from pydantic import BaseModel

from .playbook_deploy_config import PlaybookDeployConfig
from ..sinks.sink_config import SinkConfigBase


class RunnerConfig(BaseModel):
    playbook_sets: List[str] = ["defaults", "custom"]
    sinks_config: Optional[List[SinkConfigBase]] = []
    global_config: Optional[dict] = {}
    active_playbooks: Optional[List[PlaybookDeployConfig]] = []



from .sink_config import SinkConfigBase
from ...core.reporting.blocks import Finding


class SinkBase:
    def __init__(self, sink_config: SinkConfigBase):
        self.sink_name = sink_config.sink_name
        self.params = sink_config.params

    def stop(self):
        pass

    def write_finding(self, finding: Finding):
        raise NotImplementedError(
            f"write_finding not implemented for sink {self.sink_name}"
        )



import textwrap

from ...core.model.env_vars import INSTALLATION_NAMESPACE


def get_deployment_yaml(name, image="busybox"):
    return textwrap.dedent(
        f"""\
  apiVersion: apps/v1
  kind: Deployment
  metadata:
    name: {name}
    namespace: {INSTALLATION_NAMESPACE}
    labels:
      app: {name}
  spec:
    replicas: 1
    selector:
      matchLabels:
        app: {name}
    template:
      metadata:
        labels:
          app: {name}
      spec:
        containers:
        - name: runner
          image: {image}
          imagePullPolicy: Always
  """
    )



from .blocks import *
from .callbacks import SinkCallbackEvent, on_sink_callback

__all__ = [
    "BaseBlock",
    "MarkdownBlock",
    "DividerBlock",
    "FileBlock",
    "HeaderBlock",
    "ListBlock",
    "TableBlock",
    "KubernetesFieldsBlock",
    "CallbackBlock",
    "SinkCallbackEvent",
    "on_sink_callback",
]



import websocket
import json
import os
import time
from threading import Thread

from ...core.model.env_vars import TARGET_ID
from ...core.reporting.callbacks import *

SLACK_WEBSOCKET_RELAY_ADDRESS = os.environ.get("SLACK_WEBSOCKET_RELAY_ADDRESS", "wss://relay.robusta.dev")
SLACK_RECEIVER_ENABLED = os.environ.get("SLACK_RECEIVER_ENABLED", "True")
SLACK_ENABLE_WEBSOCKET_TRACING = os.environ.get("SLACK_ENABLE_WEBSOCKET_TRACING", False)
SLACK_WEBSOCKET_RECONNECT_DELAY_SEC = os.environ.get(
    "SLACK_WEBSOCKET_RECONNECT_DELAY_SEC", 3
)


def run_report_callback(action, body):
    try:
        callback_request = PlaybookCallbackRequest.parse_raw(action["value"])
        func = callback_registry.lookup_callback(callback_request)
        channel = body["channel"]["name"]
        event = SinkCallbackEvent(
            source_channel_id=body["channel"]["id"],
            source_channel_name=channel,
            source_user_id=body["user"]["id"],
            source_message=body["message"]["text"],
            source_context=callback_request.context,
        )
        logging.info(f"got callback `{func}`")
        if func is None:
            logging.error(
                f"no callback found for action_id={action['action_id']} with value={action['value']}"
            )
            return
        func(event)
        context = json.loads(callback_request.context)
        sink_name = context["sink_name"]
        # TODO Can we solve this cyclic import better?
        from ...core.sinks.sink_manager import SinkManager

        SinkManager.get_sink_by_name(sink_name).write_finding(event.finding)
    except Exception as e:
        logging.error(f"Error running callback; action={action}; e={e}")


def start_slack_receiver():
    if SLACK_RECEIVER_ENABLED != "True":
        logging.info(
            "Slack outgoing messages only mode. Slack receiver not initialized"
        )
        return

    if SLACK_WEBSOCKET_RELAY_ADDRESS == "":
        logging.warning("Slack relay adress empty. Not initializing slack relay")
        return

    websocket.enableTrace(SLACK_ENABLE_WEBSOCKET_TRACING)
    receiver_thread = Thread(target=run_forever)
    receiver_thread.start()


def run_forever():
    logging.info("starting slack relay receiver")
    while True:
        ws = websocket.WebSocketApp(
            SLACK_WEBSOCKET_RELAY_ADDRESS,
            on_open=on_open,
            on_message=on_message,
            on_error=on_error,
        )
        ws.run_forever()
        logging.info("slack relay websocket closed")
        time.sleep(SLACK_WEBSOCKET_RECONNECT_DELAY_SEC)


def on_message(ws, message):
    # TODO: use typed pydantic classes here?
    logging.debug(f"received slack message {message}")
    slack_event = json.loads(message)
    actions = slack_event["actions"]
    for action in actions:
        run_report_callback(action, slack_event)


def on_error(ws, error):
    logging.info(f"slack relay websocket error: {error}")


def on_open(ws):
    logging.info(f"connecting to server as {TARGET_ID}")
    ws.send(json.dumps({"action": "auth", "key": "dummy key", "target_id": TARGET_ID}))



# TODO: add a KubernetesBlock for rendering Kubernetes object in a standard way
# Notes on how we define all the classes below:
# 1. We use pydantic and not dataclasses so that field types are validated
# 2. We add __init__ methods ourselves for convenience. Without our own __init__ method, something like
#       HeaderBlock("foo") doesn't work. Only HeaderBlock(text="foo") would be allowed by pydantic.
import textwrap
import uuid
from copy import deepcopy
from typing import List, Callable, Dict, Any, Iterable, Sequence, Optional

import hikaru
from hikaru import DiffDetail
from hikaru.model import HikaruDocumentBase
from pydantic import BaseModel
from tabulate import tabulate
from enum import Enum

from .custom_rendering import render_value
from ..reporting.consts import FindingSource, FindingSubjectType, FindingType

BLOCK_SIZE_LIMIT = 2997  # due to slack block size limit of 3000


class BaseBlock(BaseModel):
    hidden: bool = False


class MarkdownBlock(BaseBlock):
    text: str

    def __init__(self, text: str, single_paragraph: bool = False):
        if single_paragraph:
            text = textwrap.dedent(text)
            text = text.replace("\n", "")

        if len(text) >= BLOCK_SIZE_LIMIT:
            text = text[:BLOCK_SIZE_LIMIT] + "..."
        super().__init__(text=text)


class DividerBlock(BaseBlock):
    pass


class FileBlock(BaseBlock):
    filename: str
    contents: bytes

    def __init__(self, filename: str, contents: bytes):
        super().__init__(filename=filename, contents=contents)


class HeaderBlock(BaseBlock):
    text: str

    def __init__(self, text: str):
        super().__init__(text=text)


class ListBlock(BaseBlock):
    items: List[str]

    def __init__(self, items: List[str]):
        super().__init__(items=items)

    def to_markdown(self) -> MarkdownBlock:
        mrkdwn = [f"* {item}" for item in self.items]
        return MarkdownBlock("\n".join(mrkdwn))


# TODO: we should add a generalization of this which isn't K8s specific
class KubernetesDiffBlock(BaseBlock):
    diffs: List[DiffDetail]
    old: Optional[str]
    new: Optional[str]

    # note that interesting_diffs might be a subset of the full diff between old and new
    def __init__(
        self,
        interesting_diffs: List[DiffDetail],
        old: Optional[HikaruDocumentBase],
        new: Optional[HikaruDocumentBase],
    ):
        super().__init__(
            diffs=interesting_diffs,
            old=self._obj_to_text(old),
            new=self._obj_to_text(new),
        )

    @staticmethod
    def _obj_to_text(obj: Optional[HikaruDocumentBase]):
        if obj is None:
            return ""
        else:
            return hikaru.get_yaml(obj)


class JsonBlock(BaseBlock):
    json_str: str

    def __init__(self, json_str: str):
        super().__init__(json_str=json_str)


class TableBlock(BaseBlock):
    rows: List[List]
    headers: Sequence[str] = ()
    column_renderers: Dict = {}

    def __init__(
        self, rows: List[List], headers: Sequence[str] = (), column_renderers: Dict = {}
    ):
        super().__init__(rows=rows, headers=headers, column_renderers=column_renderers)

    def to_markdown(self) -> MarkdownBlock:
        # TODO: when the next version of tabulate is released, use maxcolwidths to wrap lines that are too long
        # this is currently implemented on tabulate's git master but isn't yet in the pypi package
        # unfortunately, we can't take a dependency on the tabulate git version as that breaks our package with pypi
        # see https://github.com/python-poetry/poetry/issues/2828
        table = tabulate(self.render_rows(), headers=self.headers, tablefmt="presto")
        return MarkdownBlock(f"```\n{table}\n```")

    def render_rows(self) -> List[List]:
        if self.column_renderers is None:
            return self.rows
        new_rows = deepcopy(self.rows)
        for (column_name, renderer_type) in self.column_renderers.items():
            column_idx = self.headers.index(column_name)
            for row in new_rows:
                row[column_idx] = render_value(renderer_type, row[column_idx])
        return new_rows


class KubernetesFieldsBlock(TableBlock):
    def __init__(
        self,
        k8s_obj: HikaruDocumentBase,
        fields: List[str],
        explanations: Dict[str, str] = {},
    ):
        """
        :param k8s_obj: a kubernetes object
        :param fields: a list of fields to display. for example ["metadata.name", "metadata.namespace"]
        :param explanations: an explanation for each field. for example {"metdata.name": "the pods name"}
        """
        if explanations:
            rows = [
                [f, k8s_obj.object_at_path(f.split(".")), explanations.get(f, "")]
                for f in fields
            ]
            super().__init__(rows=rows, headers=["field", "value", "explanation"])
        else:
            rows = [[f, k8s_obj.object_at_path(f.split("."))] for f in fields]
            super().__init__(rows=rows, headers=["field", "value"])


class CallbackBlock(BaseBlock):
    choices: Dict[str, Callable]
    context: Dict[str, Any] = {}

    def __init__(self, choices: Dict[str, Callable], context: Dict[str, Any]):
        super().__init__(choices=choices, context=context)


class FindingSeverity(Enum):
    INFO = 1
    LOW = 2
    MEDIUM = 3
    HIGH = 4


class Enrichment:
    # These is the actual enrichment data
    blocks: List[BaseBlock] = []
    # General purpose rendering flags, that can be used by specific sinks
    annotations: Dict[str, str] = {}

    def __init__(self, blocks: List[BaseBlock], annotations=None):
        if annotations is None:
            annotations = {}
        self.blocks = blocks
        self.annotations = annotations

    def __str__(self):
        return f"annotations: {self.annotations} Enrichment: {self.blocks} "


class FindingSubject:
    def __init__(
        self,
        name: str = None,
        subject_type: FindingSubjectType = FindingSubjectType.TYPE_NONE,
        namespace: str = None,
    ):
        self.name = name
        self.subject_type = subject_type
        self.namespace = namespace


class Finding:
    def __init__(
        self,
        title: str,
        severity: FindingSeverity = FindingSeverity.INFO,
        source: FindingSource = FindingSource.NONE,
        aggregation_key: str = None,
        description: str = None,
        subject: FindingSubject = FindingSubject(),
        finding_type: FindingType = FindingType.ISSUE,
        failure: bool = True,
    ) -> None:
        self.id: uuid = uuid.uuid4()
        self.title = title
        self.finding_type = finding_type
        self.failure = failure
        self.description = description
        self.source = source
        self.aggregation_key = aggregation_key
        self.severity = severity
        self.category = None  # TODO fill real category
        self.subject = subject
        self.enrichments: List[Enrichment] = []

    def add_enrichment(self, enrichment_blocks: List[BaseBlock], annotations=None):
        if not enrichment_blocks:
            return
        if annotations is None:
            annotations = {}
        self.enrichments.append(Enrichment(enrichment_blocks, annotations))

    def __str__(self):
        return f"title: {self.title} desc: {self.description} severity: {self.severity} sub-name: {self.subject.name} sub-type:{self.subject.subject_type.value} enrich: {self.enrichments}"



import json
import logging
import tempfile

from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError
from hikaru import DiffType

from ...core.model.events import *
from ...core.reporting.blocks import *
from ...core.reporting.utils import add_pngs_for_all_svgs
from ...core.reporting.callbacks import PlaybookCallbackRequest
from ...core.reporting.consts import SlackAnnotations
from ...core.model.env_vars import TARGET_ID


ACTION_TRIGGER_PLAYBOOK = "trigger_playbook"
SlackBlock = Dict[str, Any]


class SlackSender:
    def __init__(self, slack_token: str):
        """
        Connect to Slack and verify that the Slack token is valid.
        Return True on success, False on failure
        """
        self.slack_client = WebClient(token=slack_token)

        try:
            self.slack_client.auth_test()
        except SlackApiError as e:
            logging.error(f"Cannot connect to Slack API: {e}")
            raise e

    @staticmethod
    def __get_action_block_for_choices(choices: Dict[str, Callable] = None, context=""):
        if choices is None:
            return []

        buttons = []
        for (i, (text, callback)) in enumerate(choices.items()):
            buttons.append(
                {
                    "type": "button",
                    "text": {
                        "type": "plain_text",
                        "text": text,
                    },
                    "style": "primary",
                    "action_id": f"{ACTION_TRIGGER_PLAYBOOK}_{i}",
                    "value": PlaybookCallbackRequest.create_for_func(
                        callback, context, text
                    ).json(),
                }
            )

        return [{"type": "actions", "elements": buttons}]

    @staticmethod
    def __apply_length_limit(msg: str, max_length: int = 3000):
        if len(msg) <= max_length:
            return msg
        truncator = "..."
        return msg[: max_length - len(truncator)] + truncator

    @staticmethod
    def __to_slack_diff(block: KubernetesDiffBlock, sink_name: str) -> List[SlackBlock]:
        # this can happen when a block.old=None or block.new=None - e.g. the resource was added or deleted
        if not block.diffs:
            return []

        num_additions = len([d for d in block.diffs if d.diff_type == DiffType.ADDED])
        num_subtractions = len(
            [d for d in block.diffs if d.diff_type == DiffType.REMOVED]
        )
        num_modifications = len(block.diffs) - num_additions - num_subtractions
        slack_blocks = []
        slack_blocks.extend(
            SlackSender.__to_slack(
                MarkdownBlock(
                    f"{num_additions} fields added. {num_subtractions} fields removed. {num_modifications} fields changed"
                ),
                sink_name,
            )
        )
        slack_blocks.extend(
            SlackSender.__to_slack(
                ListBlock(
                    [
                        f"*{d.formatted_path}*: {d.other_value} :arrow_right: {d.value}"
                        for d in block.diffs
                    ]
                ),
                sink_name,
            )
        )

        return slack_blocks

    @staticmethod
    def __to_slack(block: BaseBlock, sink_name: str) -> List[SlackBlock]:
        if isinstance(block, MarkdownBlock):
            if not block.text:
                return []
            return [
                {
                    "type": "section",
                    "text": {
                        "type": "mrkdwn",
                        "text": SlackSender.__apply_length_limit(block.text),
                    },
                }
            ]
        elif isinstance(block, DividerBlock):
            return [{"type": "divider"}]
        elif isinstance(block, FileBlock):
            raise AssertionError("to_slack() should never be called on a FileBlock")
        elif isinstance(block, HeaderBlock):
            return [
                {
                    "type": "header",
                    "text": {
                        "type": "plain_text",
                        "text": SlackSender.__apply_length_limit(block.text, 150),
                    },
                }
            ]
        elif isinstance(block, ListBlock) or isinstance(block, TableBlock):
            return SlackSender.__to_slack(block.to_markdown(), sink_name)
        elif isinstance(block, KubernetesDiffBlock):
            return SlackSender.__to_slack_diff(block, sink_name)
        elif isinstance(block, CallbackBlock):
            context = block.context.copy()
            context["target_id"] = TARGET_ID
            context["sink_name"] = sink_name
            return SlackSender.__get_action_block_for_choices(
                block.choices, json.dumps(context)
            )
        else:
            logging.error(
                f"cannot convert block of type {type(block)} to slack format block: {block}"
            )
            return []  # no reason to crash the entire report

    def __upload_file_to_slack(self, block: FileBlock) -> str:
        """Upload a file to slack and return a link to it"""
        with tempfile.NamedTemporaryFile() as f:
            f.write(block.contents)
            f.flush()
            result = self.slack_client.files_upload(
                title=block.filename, file=f.name, filename=block.filename
            )
            return result["file"]["permalink"]

    def prepare_slack_text(self, message: str, files: List[FileBlock] = []):
        if files:
            # it's a little annoying but it seems like files need to be referenced in `title` and not just `blocks`
            # in order to be actually shared. well, I'm actually not sure about that, but when I tried adding the files
            # to a separate block and not including them in `title` or the first block then the link was present but
            # the file wasn't actually shared and the link was broken
            uploaded_files = []
            for file_block in files:
                permalink = self.__upload_file_to_slack(file_block)
                uploaded_files.append(f"* <{permalink} | {file_block.filename}>")

            file_references = "\n".join(uploaded_files)
            message = f"{message}\n{file_references}"

        if len(message) == 0:
            return "empty-message"  # blank messages aren't allowed

        return SlackSender.__apply_length_limit(message)

    def __send_blocks_to_slack(
        self,
        report_blocks: List[BaseBlock],
        report_attachment_blocks: List[BaseBlock],
        title: str,
        slack_channel: str,
        unfurl: bool,
        sink_name: str,
    ):
        file_blocks = add_pngs_for_all_svgs(
            [b for b in report_blocks if isinstance(b, FileBlock)]
        )
        other_blocks = [b for b in report_blocks if not isinstance(b, FileBlock)]

        message = self.prepare_slack_text(title, file_blocks)

        output_blocks = []
        if title:
            output_blocks.extend(SlackSender.__to_slack(HeaderBlock(title), sink_name))
        for block in other_blocks:
            output_blocks.extend(SlackSender.__to_slack(block, sink_name))
        attachment_blocks = []
        for block in report_attachment_blocks:
            attachment_blocks.extend(SlackSender.__to_slack(block, sink_name))

        logging.debug(
            f"--sending to slack--\n"
            f"title:{title}\n"
            f"blocks: {output_blocks}\n"
            f"attachment_blocks: {report_attachment_blocks}\n"
            f"message:{message}"
        )

        try:
            if attachment_blocks:
                self.slack_client.chat_postMessage(
                    channel=slack_channel,
                    text=message,
                    blocks=output_blocks,
                    display_as_bot=True,
                    attachments=[{"blocks": attachment_blocks}],
                    unfurl_links=unfurl,
                    unfurl_media=unfurl,
                )
            else:
                self.slack_client.chat_postMessage(
                    channel=slack_channel,
                    text=message,
                    blocks=output_blocks,
                    display_as_bot=True,
                    unfurl_links=unfurl,
                    unfurl_media=unfurl,
                )
        except Exception as e:
            logging.error(
                f"error sending message to slack\ne={e}\ntext={message}\nblocks={output_blocks}\nattachment_blocks={attachment_blocks}"
            )

    def send_finding_to_slack(
        self, finding: Finding, slack_channel: str, sink_name: str
    ):
        blocks: List[BaseBlock] = []
        attachment_blocks: List[BaseBlock] = []
        # first add finding description block
        if finding.description:
            blocks.append(MarkdownBlock(finding.description))

        unfurl = True
        for enrichment in finding.enrichments:
            # if one of the enrichment specified unfurl=False, this slack message will contain unfurl=False
            unfurl = unfurl and enrichment.annotations.get(
                SlackAnnotations.UNFURL, True
            )
            if enrichment.annotations.get(SlackAnnotations.ATTACHMENT):
                attachment_blocks.extend(enrichment.blocks)
            else:
                blocks.extend(enrichment.blocks)

        self.__send_blocks_to_slack(
            blocks, attachment_blocks, finding.title, slack_channel, unfurl, sink_name
        )



import logging
from functools import wraps
from typing import NamedTuple, Union, List
from hikaru.model.rel_1_16 import *

from .models import PrometheusEvent, PrometheusKubernetesAlert, PrometheusAlert
from ..base_handler import handle_event
from ..kubernetes.custom_models import (
    RobustaPod,
    traceback,
    RobustaDeployment,
    RobustaJob,
)

from ...core.model.playbook_hash import playbook_hash
from ...integrations.kubernetes.base_triggers import prefix_match
from ...core.active_playbooks import register_playbook, activate_playbook
from ...core.model.trigger_params import TriggerParams
from ...core.model.cloud_event import CloudEvent
from ...core.model.events import EventType
from ...utils.decorators import doublewrap


def find_node_by_ip(ip) -> Node:
    nodes: NodeList = NodeList.listNode().obj
    for node in nodes.items:
        addresses = [a.address for a in node.status.addresses]
        logging.info(f"node {node.metadata.name} has addresses {addresses}")
        if ip in addresses:
            return node
    raise Exception(f"No node exists with IP '{ip}'")


@doublewrap
def on_pod_prometheus_alert(
    func,
    alert_name="",
    pod_name_prefix="",
    namespace_prefix="",
    instance_name_prefix="",
    status="",
):
    register_playbook(
        func,
        deploy_on_pod_prometheus_alert,
        TriggerParams(
            alert_name=alert_name,
            pod_name_prefix=pod_name_prefix,
            namespace_prefix=namespace_prefix,
            instance_name_prefix=instance_name_prefix,
            status=status,
        ),
    )
    return func


def does_alert_match_trigger(
    alert: PrometheusAlert, trigger_params: TriggerParams
) -> bool:
    if (
        trigger_params.alert_name
        and trigger_params.alert_name != alert.labels["alertname"]
    ):
        return False
    if trigger_params.status != "" and trigger_params.status != alert.status:
        return False
    if not prefix_match(trigger_params.pod_name_prefix, alert.labels.get("pod")):
        return False
    if not prefix_match(trigger_params.namespace_prefix, alert.labels.get("namespace")):
        return False
    if not prefix_match(
        trigger_params.instance_name_prefix, alert.labels.get("instance")
    ):
        return False
    return True


class ResourceMapping(NamedTuple):
    hikaru_class: Union[RobustaPod, RobustaDeployment, Job]
    attribute_name: str
    prometheus_label: str


MAPPINGS = [
    ResourceMapping(RobustaPod, "pod", "pod"),
    ResourceMapping(RobustaDeployment, "deployment", "deployment"),
    ResourceMapping(RobustaJob, "job", "job_name"),
    ResourceMapping(DaemonSet, "daemonset", "daemonset"),
]


def load_node(alert: PrometheusAlert, node_name: str) -> Node:
    node = None
    try:
        # sometimes we get an IP:PORT instead of the node name. handle that case
        if ":" in node_name:
            node = find_node_by_ip(node_name.split(":")[0])
        else:
            node = Node().read(name=node_name)
    except Exception as e:
        logging.info(f"Error loading Node kubernetes object {alert}. error: {e}")

    return node


def build_prometheus_event(alert: PrometheusAlert) -> PrometheusKubernetesAlert:
    event = PrometheusKubernetesAlert(
        alert=alert,
        alert_name=alert.labels["alertname"],
        alert_severity=alert.labels.get("severity"),
    )

    namespace = alert.labels.get("namespace", "default")

    for mapping in MAPPINGS:
        try:
            resource_name = alert.labels.get(mapping.prometheus_label, None)
            if not resource_name:
                continue
            resource = mapping.hikaru_class().read(
                name=resource_name, namespace=namespace
            )
            setattr(event, mapping.attribute_name, resource)
            logging.info(
                f"Successfully loaded Kubernetes resource {resource_name} for alert {event.alert_name}"
            )
        except Exception as e:
            logging.info(
                f"Error loading {mapping.hikaru_class} kubernetes object {event.alert}. error: {e}"
            )

    node_name = alert.labels.get("node")
    if node_name:
        event.node = load_node(alert, node_name)

    # we handle nodes differently than other resources
    node_name = event.alert.labels.get("instance", None)
    job_name = event.alert.labels.get(
        "job", None
    )  # a prometheus "job" not a kubernetes "job" resource
    # when the job_name is kube-state-metrics "instance" refers to the IP of kube-state-metrics not the node
    if not event.node and node_name and job_name != "kube-state-metrics":
        event.node = load_node(alert, node_name)

    return event


def handle_single_alert(
    alert: PrometheusAlert,
    trigger_params: TriggerParams,
    named_sinks: List[str],
    func,
    action_params=None,
):
    if not does_alert_match_trigger(alert, trigger_params):
        return

    event = build_prometheus_event(alert)

    return handle_event(func, event, action_params, "prometheus", named_sinks)


def deploy_on_pod_prometheus_alert(
    func, trigger_params: TriggerParams, named_sinks: List[str], action_params=None
):
    @wraps(func)
    def wrapper(cloud_event: CloudEvent):
        logging.debug(
            f"checking if we should run {func} on prometheus event {trigger_params.alert_name}"
        )
        prometheus_event = PrometheusEvent(**cloud_event.data)
        results = []
        for alert in prometheus_event.alerts:
            try:
                result = handle_single_alert(
                    alert, trigger_params, named_sinks, func, action_params
                )
                if result:
                    results.append(result)
            except Exception:
                logging.error(
                    f"Failed to process alert {alert} {traceback.format_exc()}"
                )

        return ",".join(results)

    playbook_id = playbook_hash(func, trigger_params, action_params)
    activate_playbook(EventType.PROMETHEUS, wrapper, func, playbook_id)
    return wrapper



import base64
import os
import time
import uuid

import requests
import typer
from pydantic import BaseModel

from .utils import log_title

app = typer.Typer()

SLACK_INTEGRATION_SERVICE_ADDRESS = os.environ.get(
    "SLACK_INTEGRATION_SERVICE_ADDRESS",
    "https://robusta.dev/integrations/slack/get-token",
)


def wait_for_slack_api_key(id: str) -> str:
    while True:
        try:
            response_json = requests.get(
                f"{SLACK_INTEGRATION_SERVICE_ADDRESS}?id={id}"
            ).json()
            if response_json["token"]:
                return str(response_json["token"])
            time.sleep(0.5)
        except Exception as e:
            log_title(f"Error getting slack token {e}")


def get_slack_key():
    id = str(uuid.uuid4())
    url = f"https://robusta.dev/integrations/slack?id={id}"
    typer.secho(
        f"If your browser does not automatically launch, open the below url:\n{url}"
    )
    typer.launch(url)
    slack_api_key = wait_for_slack_api_key(id)
    return slack_api_key


@app.command()
def slack():
    """generate slack api key"""
    key = get_slack_key()
    log_title(f"your slack key is:\n{key}\nAdd it to the slack sink configuration")


if __name__ == "__main__":
    app()



from collections import OrderedDict
from hikaru.model import Node
from prometheus_api_client import PrometheusConnect
from ..prometheus.utils import PrometheusDiscovery


class NodeAnalyzer:

    # TODO: perhaps we should handle this more elegantly by first loading all the data into a pandas dataframe
    # and then slicing it different ways
    def __init__(self, node: Node, prometheus_url: str, range_size="5m"):
        self.node = node
        self.range_size = range_size
        self.internal_ip = next(
            addr.address
            for addr in self.node.status.addresses
            if addr.type == "InternalIP"
        )
        if prometheus_url is None:
            prometheus_url = PrometheusDiscovery.find_prometheus_url()

        self.prom = PrometheusConnect(url=prometheus_url, disable_ssl=True)

    def get_total_cpu_usage(self, other_method=False):
        """
        Gets the total cpu usage for the node, including both containers and everything running on the host directly
        :return: a float between 0 and 1 representing the percentage of total cpus used
        """
        if other_method:
            return self._query(
                f'rate(container_cpu_usage_seconds_total{{node="{self.node.metadata.name}",pod="",id="/"}}[{self.range_size}]) '
                f'/ scalar(sum (machine_cpu_cores{{node="{self.node.metadata.name}"}}))'
            )

        # the instance here refers to the node as identified by it's internal IP
        # we average by the instance to account for multiple cpus and still return a number between 0-1
        return self._query(
            f"1"
            f"- avg by(instance)(rate("
            f'   node_cpu_seconds_total{{mode=~"idle", instance=~"{self.node.metadata.name}|{self.internal_ip}:.*"}}[{self.range_size}]'
            f"))"
            f"- avg by(instance)(rate("
            f'   node_cpu_seconds_total{{mode=~"iowait", instance=~"{self.node.metadata.name}|{self.internal_ip}:.*"}}[{self.range_size}]'
            f"))"
        )

    def get_total_containerized_cpu_usage(self):
        query = self._build_query_for_containerized_cpu_usage(True, True)
        return self._query(query)

    def get_per_pod_cpu_usage(self, threshold=0.0, normalize_by_cpu_count=True):
        """
        Gets the cpu usage of each pod on a node
        :param threshold: only return pods with a cpu above threshold
        :param normalize_by_cpu_count: should we divide by the number of cpus so that the result is in the range 0-1 regardless of cpu count?
        :return: a dict of {[pod_name] : [cpu_usage in the 0-1 range] }
        """
        query = self._build_query_for_containerized_cpu_usage(
            False, normalize_by_cpu_count
        )
        result = self.prom.custom_query(query)
        pod_value_pairs = [(r["metric"]["pod"], float(r["value"][1])) for r in result]
        pod_value_pairs = [(k, v) for (k, v) in pod_value_pairs if v >= threshold]
        pod_value_pairs.sort(key=lambda x: x[1], reverse=True)
        pod_to_cpu = OrderedDict(pod_value_pairs)
        return pod_to_cpu

    def get_per_pod_cpu_request(self):
        query = f'sum by (pod)(kube_pod_container_resource_requests_cpu_cores{{node="{self.node.metadata.name}"}})'
        result = self.prom.custom_query(query)
        return dict((r["metric"]["pod"], float(r["value"][1])) for r in result)

    def _query(self, query):
        """
        Runs a simple query returning a single metric and returns that metric
        """
        result = self.prom.custom_query(query)
        return float(result[0]["value"][1])

    def _build_query_for_containerized_cpu_usage(self, total, normalized_by_cpu_count):
        if total:
            grouping = ""
        else:
            grouping = "by (pod)"

        if normalized_by_cpu_count:
            # we divide by the number of machine_cpu_cores to return a result in th 0-1 range regardless of cpu count
            normalization = (
                f'/ scalar(sum (machine_cpu_cores{{node="{self.node.metadata.name}"}}))'
            )
        else:
            normalization = ""

        # note: it is important to set either image!="" or image="" because otherwise we count everything twice -
        # once for the whole pod (image="") and once for each container (image!="")
        return (
            f"sum(rate("
            f'           container_cpu_usage_seconds_total{{node="{self.node.metadata.name}",pod!="",image!=""}}[{self.range_size}]'
            f")) {grouping} {normalization}"
        )
